"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides some utilities related to the sorting mechanism.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from typing import Any, Callable, List, Union

import PyFunceble.facility
import PyFunceble.storage
from PyFunceble.converter.url2netloc import Url2Netloc
from PyFunceble.helpers.regex import RegexHelper


def standard(element: Any) -> List[Union[int, Any]]:
    """
    Provides the key to use for the standard sorting.

    :param element:
        The element to format.
    """

    element = element.strip()

    if not element:
        return []

    regex_helper = RegexHelper()

    element = Url2Netloc(element).get_converted().strip()

    if PyFunceble.facility.ConfigLoader.is_already_loaded():
        element = regex_helper.set_regex(
            r"^%s\s+" % PyFunceble.storage.CONFIGURATION.cli_testing.hosts_ip
        ).replace_match(element, "")

    cleaned = regex_helper.set_regex(r"[^a-zA-Z0-9\.]").replace_match(element, "")

    return [
        int(x) if x.isdigit() else x
        for x in regex_helper.set_regex(r"(\d+)").split(cleaned)
    ]


def hierarchical(element: Any) -> List[Union[int, Any]]:
    """
    Provides the key to use for the hierarchical sorting.

    :param element:
        The element to format.
    """

    element = element.strip()

    if not element:
        return []

    element = Url2Netloc(element).get_converted().strip()

    if PyFunceble.facility.ConfigLoader.is_already_loaded():
        element = RegexHelper(
            r"^%s\s+" % PyFunceble.storage.CONFIGURATION.cli_testing.hosts_ip
        ).replace_match(element, "")

    return standard(".".join(reversed(element.split("."))))


def get_best_sorting_key() -> Callable[[Any], List[Union[int, Any]]]:
    """
    Provides the best sorting key from the configuration.
    """

    if PyFunceble.facility.ConfigLoader.is_already_loaded():
        if PyFunceble.storage.CONFIGURATION.cli_testing.sorting_mode.hierarchical:
            return hierarchical
    return standard



"""Base

Revision ID: ade87195b0a0
Revises:
Create Date: 2020-08-15 20:12:26.768419

"""
import sqlalchemy as sa
from alembic import op

# pylint: skip-file

# revision identifiers, used by Alembic.
revision = "ade87195b0a0"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "pyfunceble_file",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.Column("modified", sa.DateTime(), nullable=True),
        sa.Column("path", sa.Text(), nullable=False),
        sa.Column("test_completed", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("path"),
    )
    op.create_table(
        "pyfunceble_whois_record",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.Column("modified", sa.DateTime(), nullable=True),
        sa.Column("subject", sa.Text(), nullable=False),
        sa.Column("expiration_date", sa.Text(), nullable=False),
        sa.Column("epoch", sa.Integer(), nullable=False),
        sa.Column("state", sa.String(length=80), nullable=False),
        sa.Column("record", sa.Text(), nullable=False),
        sa.Column("server", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("subject"),
    )
    op.create_table(
        "pyfunceble_status",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.Column("modified", sa.DateTime(), nullable=True),
        sa.Column("file_id", sa.Integer(), nullable=False),
        sa.Column("tested", sa.Text(), nullable=False),
        sa.Column("_status", sa.Text(), nullable=True),
        sa.Column("status", sa.Text(), nullable=True),
        sa.Column("_status_source", sa.Text(), nullable=True),
        sa.Column("status_source", sa.Text(), nullable=True),
        sa.Column("domain_syntax_validation", sa.Boolean(), nullable=True),
        sa.Column("expiration_date", sa.Text(), nullable=True),
        sa.Column("http_status_code", sa.Integer(), nullable=True),
        sa.Column("ipv4_range_syntax_validation", sa.Boolean(), nullable=True),
        sa.Column("ipv4_syntax_validation", sa.Boolean(), nullable=True),
        sa.Column("ipv6_range_syntax_validation", sa.Boolean(), nullable=True),
        sa.Column("ipv6_syntax_validation", sa.Boolean(), nullable=True),
        sa.Column("subdomain_syntax_validation", sa.Boolean(), nullable=True),
        sa.Column("url_syntax_validation", sa.Boolean(), nullable=True),
        sa.Column("is_complement", sa.Boolean(), nullable=True),
        sa.Column("test_completed", sa.Boolean(), nullable=False),
        sa.Column("tested_at", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(
            ["file_id"], ["pyfunceble_file.id"], onupdate="CASCADE", ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("tested"),
    )
    op.create_table(
        "pyfunceble_mined",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.Column("modified", sa.DateTime(), nullable=True),
        sa.Column("subject_id", sa.Integer(), nullable=False),
        sa.Column("file_id", sa.Integer(), nullable=False),
        sa.Column("mined", sa.Text(), nullable=False),
        sa.ForeignKeyConstraint(
            ["file_id"], ["pyfunceble_file.id"], onupdate="CASCADE", ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["subject_id"],
            ["pyfunceble_status.id"],
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("mined"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("pyfunceble_mined")
    op.drop_table("pyfunceble_status")
    op.drop_table("pyfunceble_whois_record")
    op.drop_table("pyfunceble_file")
    # ### end Alembic commands ###



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides an easy way to convert a digit string to 2 digits.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
# pylint: enable=line-too-long

from typing import Any

from PyFunceble.query.whois.converter.base import ConverterBase


class Digit2Digits(ConverterBase):
    """
    Converts a given digit to a 2 digits string.
    """

    @ConverterBase.data_to_convert.setter
    def data_to_convert(self, value: Any) -> Any:
        """
        Overrites the default behavior.

        :param value:
            The data to convert.

        :raise TypeError:
            When the given data to convert is not :py:class:`str`
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        # pylint: disable=no-member
        super(Digit2Digits, self.__class__).data_to_convert.fset(self, value)

    @ConverterBase.ensure_data_to_convert_is_given
    def get_converted(self) -> str:
        """
        Provides the converted data (after conversion)
        """

        return str(self.data_to_convert).zfill(2)



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides some testing related utilities

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import os
from typing import List, Optional, Union

from sqlalchemy.orm import Session

import PyFunceble.storage
from PyFunceble.converter.adblock_input_line2subject import AdblockInputLine2Subject
from PyFunceble.converter.cidr2subject import CIDR2Subject
from PyFunceble.converter.input_line2subject import InputLine2Subject
from PyFunceble.converter.rpz_input_line2subject import RPZInputLine2Subject
from PyFunceble.converter.rpz_policy2subject import RPZPolicy2Subject
from PyFunceble.converter.subject2complements import Subject2Complements
from PyFunceble.converter.url2netloc import Url2Netloc
from PyFunceble.converter.wildcard2subject import Wildcard2Subject
from PyFunceble.dataset.autocontinue.csv import CSVContinueDataset
from PyFunceble.dataset.autocontinue.mariadb import MariaDBContinueDataset
from PyFunceble.dataset.autocontinue.mysql import MySQLContinueDataset
from PyFunceble.dataset.base import DatasetBase
from PyFunceble.dataset.csv_base import CSVDatasetBase
from PyFunceble.dataset.db_base import DBDatasetBase
from PyFunceble.dataset.inactive.csv import CSVInactiveDataset
from PyFunceble.dataset.inactive.mariadb import MariaDBInactiveDataset
from PyFunceble.dataset.inactive.mysql import MySQLInactiveDataset
from PyFunceble.helpers.list import ListHelper
from PyFunceble.helpers.regex import RegexHelper


def get_testing_mode() -> str:
    """
    Tries to provides the testing mode to apply to the CLI.
    """

    if PyFunceble.storage.CONFIGURATION.cli_testing.testing_mode.syntax:
        return "SYNTAX"
    if PyFunceble.storage.CONFIGURATION.cli_testing.testing_mode.reputation:
        return "REPUTATION"

    if PyFunceble.storage.CONFIGURATION.cli_testing.testing_mode.availability:
        return "AVAILABILITY"
    return "UNKNOWN"


def get_continue_databaset_object(
    db_session: Optional[Session] = None,
) -> Union[DatasetBase, CSVDatasetBase, DBDatasetBase]:
    """
    Provides the continue object to work with.

    :param db_session:
        A database session to use.

    :raise ValueError:
        When the given database type is unkown.
    """

    result = None

    if PyFunceble.storage.CONFIGURATION.cli_testing.db_type in "csv":
        result = CSVContinueDataset()
    elif PyFunceble.storage.CONFIGURATION.cli_testing.db_type == "mariadb":
        result = MariaDBContinueDataset(db_session=db_session)
    elif PyFunceble.storage.CONFIGURATION.cli_testing.db_type == "mysql":
        result = MySQLContinueDataset(db_session=db_session)

    if result:
        result.set_authorized(
            bool(PyFunceble.storage.CONFIGURATION.cli_testing.autocontinue)
        )

        return result

    raise ValueError(
        "<config.db_type> "
        f"({PyFunceble.storage.CONFIGURATION.cli_testing.db_type}) is unknown."
    )


def get_inactive_dataset_object(
    db_session: Optional[Session] = None,
) -> Union[DatasetBase, CSVDatasetBase, DBDatasetBase]:
    """
    Provides the inactive object to work with.

    :param db_session:
        A database session to use.

    :raise ValueError:
        When the given database type is unkown.
    """

    result = None

    if PyFunceble.storage.CONFIGURATION.cli_testing.db_type == "csv":
        result = CSVInactiveDataset()
    elif PyFunceble.storage.CONFIGURATION.cli_testing.db_type == "mariadb":
        result = MariaDBInactiveDataset(db_session=db_session)
    elif PyFunceble.storage.CONFIGURATION.cli_testing.db_type == "mysql":
        result = MySQLInactiveDataset(db_session=db_session)

    if result:
        result.set_authorized(
            bool(PyFunceble.storage.CONFIGURATION.cli_testing.inactive_db)
        )

        return result

    raise ValueError(
        "<config.db_type> "
        f"({PyFunceble.storage.CONFIGURATION.cli_testing.db_type}) is unknown."
    )


def get_destination_from_origin(origin: str) -> str:
    """
    Given the origin, we provides the destination.
    """

    if "/" in origin:
        origin = origin.rsplit("/", 1)[-1]

    if os.sep in origin:
        origin = origin.rsplit(os.sep, 1)[-1]

    return RegexHelper("[^a-zA-Z0-9._-]").replace_match(origin, "_")


def get_subjects_from_line(
    line: str,
    checker_type: str,
    *,
    adblock_inputline2subject: Optional[AdblockInputLine2Subject] = None,
    wildcard2subject: Optional[Wildcard2Subject] = None,
    rpz_policy2subject: Optional[RPZPolicy2Subject] = None,
    rpz_inputline2subject: Optional[RPZInputLine2Subject] = None,
    inputline2subject: Optional[InputLine2Subject] = None,
    subject2complements: Optional[Subject2Complements] = None,
    url2netloc: Optional[Url2Netloc] = None,
    cidr2subject: Optional[CIDR2Subject] = None,
) -> List[str]:
    """
    Provides the list of subject to test.
    """

    result = []

    if adblock_inputline2subject is None:
        adblock_inputline2subject = AdblockInputLine2Subject()

    if wildcard2subject is None:
        wildcard2subject = Wildcard2Subject()

    if rpz_policy2subject is None:
        rpz_policy2subject = RPZPolicy2Subject()

    if rpz_inputline2subject is None:
        rpz_inputline2subject = RPZInputLine2Subject()

    if inputline2subject is None:
        inputline2subject = InputLine2Subject()

    if subject2complements is None:
        subject2complements = Subject2Complements()

    if url2netloc is None:
        url2netloc = Url2Netloc()

    if cidr2subject is None:
        cidr2subject = CIDR2Subject()

    if PyFunceble.storage.CONFIGURATION.cli_decoding.adblock:
        result.extend(
            # pylint: disable=line-too-long
            adblock_inputline2subject.set_aggressive(
                bool(PyFunceble.storage.CONFIGURATION.cli_decoding.adblock_aggressive)
            )
            .set_data_to_convert(line)
            .get_converted()
        )
    elif PyFunceble.storage.CONFIGURATION.cli_decoding.wildcard:
        result.append(wildcard2subject.set_data_to_convert(line).get_converted())
    elif PyFunceble.storage.CONFIGURATION.cli_decoding.rpz:
        result.extend(
            [
                rpz_policy2subject.set_data_to_convert(x).get_converted()
                for x in rpz_inputline2subject.set_data_to_convert(line).get_converted()
            ]
        )
    else:
        result.extend(inputline2subject.set_data_to_convert(line).get_converted())

    if PyFunceble.storage.CONFIGURATION.cli_testing.complements:
        result.extend(
            [
                y
                for x in result
                for y in subject2complements.set_data_to_convert(x).get_converted()
            ]
        )

    if PyFunceble.storage.CONFIGURATION.cli_testing.cidr_expand:
        result = [
            y
            for x in result
            for y in cidr2subject.set_data_to_convert(x).get_converted()
        ]

    if checker_type.lower() != "syntax":
        for index, subject in enumerate(result):
            if not subject:
                continue

            netloc = url2netloc.set_data_to_convert(subject).get_converted()

            result[index] = subject.replace(netloc, netloc.lower())

    return ListHelper(result).remove_duplicates().remove_empty().subject



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides an interface for the backup of the directory structure.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import os

import PyFunceble.facility
from PyFunceble.cli.filesystem.dir_structure.base import DirectoryStructureBase
from PyFunceble.helpers.dict import DictHelper
from PyFunceble.helpers.directory import DirectoryHelper
from PyFunceble.helpers.file import FileHelper
from PyFunceble.helpers.hash import HashHelper


class DirectoryStructureBackup(DirectoryStructureBase):
    """
    Provides the base of all dir structure classes.
    """

    def get_backup_data(self) -> dict:
        """
        Provides the data which acts as a backup.
        """

        result = {}
        base_dir = self.get_output_basedir()
        file_helper = FileHelper()
        hash_helper = HashHelper()

        for file in DirectoryHelper(base_dir).list_all_files():
            file_helper.set_path(file)
            reduced_path = self.get_path_without_base_dir(file)

            directory, filename = reduced_path.rsplit(os.sep, 1)
            file_hash = hash_helper.hash_file(file_helper.path)

            dataset = {filename: {"content": file_helper.read(), "hash": file_hash}}

            if directory not in result:
                result[directory] = dataset
            else:
                result[directory].update(dataset)

        PyFunceble.facility.Logger.debug("Dir Structure to backup:\n%r", result)

        return result

    def store_backup(self) -> "DirectoryStructureBackup":
        """
        Stores the backup at the current destination.
        """

        DictHelper(self.get_backup_data()).to_json_file(self.source_file)

        PyFunceble.facility.Logger.info("Stored backup into: %r", self.source_file)

        return self

    def start(self) -> "DirectoryStructureBackup":
        """
        Starts the backup process.
        """

        return self.cleanup().store_backup()



"""Rename created to created_at and modified_to_modified_at

Revision ID: 7bcf7fa64ba1
Revises: 83ada95132bf
Create Date: 2020-12-08 17:34:59.349943

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import mysql

# pylint: skip-file

# revision identifiers, used by Alembic.
revision = "7bcf7fa64ba1"
down_revision = "83ada95132bf"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "pyfunceble_status", sa.Column("created_at", sa.DateTime(), nullable=False)
    )
    op.add_column(
        "pyfunceble_status", sa.Column("modified_at", sa.DateTime(), nullable=True)
    )
    op.drop_column("pyfunceble_status", "created")
    op.drop_column("pyfunceble_status", "modified")
    op.add_column(
        "pyfunceble_whois_record",
        sa.Column("created_at", sa.DateTime(), nullable=False),
    )
    op.add_column(
        "pyfunceble_whois_record",
        sa.Column("modified_at", sa.DateTime(), nullable=True),
    )
    op.drop_column("pyfunceble_whois_record", "created")
    op.drop_column("pyfunceble_whois_record", "modified")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "pyfunceble_whois_record",
        sa.Column("modified", mysql.DATETIME(), nullable=True),
    )
    op.add_column(
        "pyfunceble_whois_record",
        sa.Column("created", mysql.DATETIME(), nullable=False),
    )
    op.drop_column("pyfunceble_whois_record", "modified_at")
    op.drop_column("pyfunceble_whois_record", "created_at")
    op.add_column(
        "pyfunceble_status", sa.Column("modified", mysql.DATETIME(), nullable=True)
    )
    op.add_column(
        "pyfunceble_status", sa.Column("created", mysql.DATETIME(), nullable=False)
    )
    op.drop_column("pyfunceble_status", "modified_at")
    op.drop_column("pyfunceble_status", "created_at")
    op.create_table(
        "pyfunceble_file",
        sa.Column(
            "id", mysql.INTEGER(display_width=11), autoincrement=True, nullable=False
        ),
        sa.Column("created", mysql.DATETIME(), nullable=False),
        sa.Column("modified", mysql.DATETIME(), nullable=True),
        sa.Column("path", mysql.TEXT(collation="utf8mb4_unicode_ci"), nullable=False),
        sa.Column(
            "test_completed",
            mysql.TINYINT(display_width=1),
            autoincrement=False,
            nullable=False,
        ),
        sa.CheckConstraint("`test_completed` in (0,1)", name="CONSTRAINT_1"),
        sa.PrimaryKeyConstraint("id"),
        mysql_collate="utf8mb4_unicode_ci",
        mysql_default_charset="utf8mb4",
        mysql_engine="InnoDB",
    )
    # ### end Alembic commands ###



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides an interface which let us interact with the Public Suffix List.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import os
from typing import Any, List

import PyFunceble.storage
from PyFunceble.dataset.base import DatasetBase
from PyFunceble.downloader.public_suffix import PublicSuffixDownloader


class PublicSuffixDataset(DatasetBase):
    """
    Provides the dataset handler for the Public Suffix List dataset.
    """

    STORAGE_INDEX: str = "PUBLIC_SUFFIX"
    DOWNLOADER: PublicSuffixDownloader = PublicSuffixDownloader()

    def __init__(self) -> None:
        self.source_file = os.path.join(
            PyFunceble.storage.CONFIG_DIRECTORY,
            PyFunceble.storage.PUBLIC_SUFFIX_DUMP_FILENAME,
        )

    def __contains__(self, value: Any) -> bool:
        if value.startswith("."):
            value = value[1:]

        return value in self.get_content()

    def __getattr__(self, value: Any) -> List[str]:
        if value.startswith("."):
            value = value[1:]

        if value in self:
            return self.get_content()[value]

        return []

    def __getitem__(self, value: Any) -> List[str]:
        return self.__getattr__(value)

    def is_extension(self, extension: str) -> bool:
        """
        Checks if the given extension is registered.

        :raise TypeError:
            When :code:`extension` is not a :py:class:`str`.
        """

        if not isinstance(extension, str):
            raise TypeError(f"<extension> should be {str}, {type(extension)} given.")

        return extension in self

    def get_available_suffix(self, extension: str) -> List[str]:
        """
        Provides the available suffix for the extension.
        """

        if self.is_extension(extension):
            return self[extension]
        return []



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides everything related to our status file generation.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import functools
import os
from typing import Optional, Union

import PyFunceble.cli.storage
import PyFunceble.facility
import PyFunceble.storage
from PyFunceble.checker.availability.status import AvailabilityCheckerStatus
from PyFunceble.checker.reputation.status import ReputationCheckerStatus
from PyFunceble.checker.status_base import CheckerStatusBase
from PyFunceble.checker.syntax.status import SyntaxCheckerStatus
from PyFunceble.cli.filesystem.dir_base import FilesystemDirBase
from PyFunceble.cli.filesystem.dir_structure.restore import (
    DirectoryStructureRestoration,
)
from PyFunceble.cli.filesystem.printer.file import FilePrinter
from PyFunceble.helpers.directory import DirectoryHelper
from PyFunceble.utils.platform import PlatformUtility


class StatusFileGenerator(FilesystemDirBase):
    """
    Provides an interface for the generation of the status file from
    a given status.
    """

    # pylint: disable=too-many-public-methods

    STD_HOSTS_IP: str = "0.0.0.0"
    STD_ALLOW_HOSTS_FILES: bool = True
    STD_ALLOW_PLAIN_FILES: bool = True
    STD_ALLOW_ANALYTIC_FILES: bool = True
    STD_ALLOW_UNIFIED_FILE: bool = False

    file_printer: FilePrinter = FilePrinter()

    _test_dataset: Optional[dict] = None
    _status: Optional[
        Union[SyntaxCheckerStatus, AvailabilityCheckerStatus, ReputationCheckerStatus]
    ] = None

    _allow_hosts_files: bool = True
    _allow_plain_files: bool = True
    _allow_analytic_files: bool = True
    _allow_unified_file: bool = False

    _hosts_ip: Optional[str] = None

    def __init__(
        self,
        status: Optional[
            Union[
                SyntaxCheckerStatus, AvailabilityCheckerStatus, ReputationCheckerStatus
            ]
        ] = None,
        *,
        allow_hosts_files: Optional[bool] = None,
        allow_plain_files: Optional[bool] = None,
        allow_analytic_files: Optional[bool] = None,
        hosts_ip: Optional[str] = None,
        allow_unified_file: Optional[bool] = None,
        parent_dirname: Optional[str] = None,
        test_dataset: Optional[dict] = None,
    ) -> None:
        if status is not None:
            self.status = status

        if allow_hosts_files is not None:
            self.allow_hosts_files = allow_hosts_files
        else:
            self.guess_and_set_allow_hosts_files()

        if allow_plain_files is not None:
            self.allow_plain_files = allow_plain_files
        else:
            self.guess_and_set_allow_plain_files()

        if allow_analytic_files is not None:
            self.allow_analytic_files = allow_analytic_files
        else:
            self.guess_and_set_allow_analytic_files()

        if hosts_ip is not None:
            self.hosts_ip = hosts_ip
        else:
            self.guess_and_set_hosts_ip()

        if allow_unified_file is not None:
            self.allow_unified_file = allow_unified_file
        else:
            self.guess_and_set_allow_unified_file()

        if test_dataset is not None:
            self.test_dataset = test_dataset

        super().__init__(parent_dirname=parent_dirname)

    def ensure_status_is_given(func):  # pylint: disable=no-self-argument
        """
        Ensures that the status is given before launching the decorated method.

        :raise TypeError:
            When :code:`self.status` is not set.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if not isinstance(
                self.status,
                CheckerStatusBase,
            ):
                raise TypeError("<self.status> is not given.")

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    @property
    def status(self) -> CheckerStatusBase:
        """
        Provides the current state of the :code:`_status` attribute.
        """

        return self._status

    @status.setter
    def status(self, value: CheckerStatusBase) -> None:
        """
        Sets the status to work with.

        :param value:
            The value to set.

        :raise TypeError:
            When the given value is not a
            :class`~ PyFunceble.checker.syntax.status.SyntaxCheckerStatus`,
            :class:`~PyFunceble.checker.availability.status.AvailabilityCheckerStatus`
            or :class:`~PyFunceble.checker.reputation.status.ReputationCheckerStatus`
        """

        if not isinstance(
            value,
            CheckerStatusBase,
        ):
            raise TypeError(
                f"<value> should be {CheckerStatusBase}, {type(value)} given."
            )

        self._status = value

    def set_status(
        self,
        value: CheckerStatusBase,
    ) -> "StatusFileGenerator":
        """
        Sets the status to work with.

        :param value:
            The value to set.
        """

        self.status = value

        return self

    @property
    def test_dataset(self) -> Optional[dict]:
        """
        Provides the current state of the :code:`_test_dataset` attribute.
        """

        return self._test_dataset

    @test_dataset.setter
    def test_dataset(self, value: dict) -> None:
        """
        Sets the test dataset which was given to the tester.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class`dict`.
        """

        if not isinstance(value, dict):
            raise TypeError(f"<value> should be {dict}, {type(value)} given.")

        self._test_dataset = value

    def set_test_dataset(self, value: dict) -> "StatusFileGenerator":
        """
        Sets the test dataset which was given to the tester.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class`dict`.
        """

        self.test_dataset = value

        return self

    @property
    def allow_hosts_files(self) -> bool:
        """
        Provides the current state of the :code:`_allow_hosts_files` attribute.
        """

        return self._allow_hosts_files

    @allow_hosts_files.setter
    def allow_hosts_files(self, value: bool) -> None:
        """
        Sets the authorization to generation of hosts files.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`bool`.
        """

        if not isinstance(value, bool):
            raise TypeError(f"<value> should be {bool}, {type(value)} given.")

        self._allow_hosts_files = value

    def set_allow_hosts_files(self, value: bool) -> "StatusFileGenerator":
        """
        Sets the authorization to generation of hosts files.

        :param value:
            The value to set.
        """

        self.allow_hosts_files = value

        return self

    def guess_and_set_allow_hosts_files(self) -> "StatusFileGenerator":
        """
        Tries to guess the value of the :code:`allow_hosts_files` from the
        configuration file.
        """

        if PyFunceble.facility.ConfigLoader.is_already_loaded():
            self.allow_hosts_files = (
                PyFunceble.storage.CONFIGURATION.cli_testing.file_generation.hosts
            )
        else:
            self.allow_hosts_files = self.STD_ALLOW_HOSTS_FILES

    @property
    def allow_plain_files(self) -> bool:
        """
        Provides the current state of the :code:`_allow_plain_file` attribute.
        """

        return self._allow_plain_files

    @allow_plain_files.setter
    def allow_plain_files(self, value: bool) -> None:
        """
        Sets the authorization to generation of plain files.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`bool`.
        """

        if not isinstance(value, bool):
            raise TypeError(f"<value> should be {bool}, {type(value)} given.")

        self._allow_plain_files = value

    def set_allow_plain_files(self, value: bool) -> "StatusFileGenerator":
        """
        Sets the authorization to generation of plain files.

        :param value:
            The value to set.
        """

        self.allow_plain_files = value

        return self

    def guess_and_set_allow_plain_files(self) -> "StatusFileGenerator":
        """
        Tries to guess the value of the :code:`allow_plain_files` from the
        configuration file.
        """

        if PyFunceble.facility.ConfigLoader.is_already_loaded():
            self.allow_plain_files = (
                PyFunceble.storage.CONFIGURATION.cli_testing.file_generation.plain
            )
        else:
            self.allow_plain_files = self.STD_ALLOW_PLAIN_FILES

    @property
    def allow_analytic_files(self) -> bool:
        """
        Provides the current state of the :code:`_allow_analytic_files`
        attribute.
        """

        return self._allow_analytic_files

    @allow_analytic_files.setter
    def allow_analytic_files(self, value: bool) -> None:
        """
        Sets the authorization to generation of analytic files.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`bool`.
        """

        if not isinstance(value, bool):
            raise TypeError(f"<value> should be {bool}, {type(value)} given.")

        self._allow_analytic_files = value

    def set_allow_analytic_files(self, value: bool) -> "StatusFileGenerator":
        """
        Sets the authorization to generation of analytic files.

        :param value:
            The value to set.
        """

        self.allow_analytic_files = value

        return self

    def guess_and_set_allow_analytic_files(self) -> "StatusFileGenerator":
        """
        Tries to guess the value of the :code:`allow_analytic_files` from the
        configuration file.
        """

        if PyFunceble.facility.ConfigLoader.is_already_loaded():
            self.allow_analytic_files = (
                PyFunceble.storage.CONFIGURATION.cli_testing.file_generation.analytic
            )
        else:
            self.allow_analytic_files = self.STD_ALLOW_ANALYTIC_FILES

    @property
    def hosts_ip(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_hosts_ip` attribute.
        """

        return self._hosts_ip

    @hosts_ip.setter
    def hosts_ip(self, value: str) -> None:
        """
        Sets the hosts IP to use while generating the hosts files.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise ValueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empty.")

        self._hosts_ip = value

    def set_hosts_ip(self, value: str) -> "StatusFileGenerator":
        """
        Sets the hosts IP to use while generating the hosts files.

        :param value:
            The value to set.
        """

        self.hosts_ip = value

        return self

    def guess_and_set_hosts_ip(self) -> "StatusFileGenerator":
        """
        Tries to guess the value of the :code:`hosts_ip` from the
        configuration file.
        """

        if PyFunceble.facility.ConfigLoader.is_already_loaded():
            self.hosts_ip = PyFunceble.storage.CONFIGURATION.cli_testing.hosts_ip
        else:
            self.hosts_ip = self.STD_HOSTS_IP

    @property
    def allow_unified_file(self) -> bool:
        """
        Provides the current state of the :code:`allow_unified_file` attribute.
        """

        return self._allow_unified_file

    @allow_unified_file.setter
    def allow_unified_file(self, value: bool) -> None:
        """
        Sets the authorization to generation of the unified status file.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`bool`.
        """

        if not isinstance(value, bool):
            raise TypeError(f"<value> should be {bool}, {type(value)} given.")

        self._allow_unified_file = value

    def set_allow_unified_file(self, value: bool) -> "StatusFileGenerator":
        """
        Sets the authorization to generation of the unified status file.

        :param value:
            The value to set.
        """

        self.allow_unified_file = value

        return self

    def guess_and_set_allow_unified_file(self) -> "StatusFileGenerator":
        """
        Tries to guess the value of the :code:`allow_unified_file` from the
        configuration file.
        """

        if PyFunceble.facility.ConfigLoader.is_already_loaded():
            # pylint: disable=line-too-long
            self.allow_unified_file = (
                PyFunceble.storage.CONFIGURATION.cli_testing.file_generation.unified_results
            )
        else:
            self.allow_unified_file = self.STD_ALLOW_UNIFIED_FILE

    def guess_all_settings(self) -> "StatusFileGenerator":
        """
        Try to guess all settings.
        """

        to_ignore = ["guess_all_settings"]

        for method in dir(self):
            if method in to_ignore or not method.startswith("guess_"):
                continue

            getattr(self, method)()

        return self

    def get_output_basedir(self) -> str:
        """
        Provides the output base directory.

        :param create_if_missing:
            Authorizes the creation of the directory if it's missing.
        """

        result = super().get_output_basedir()

        if not DirectoryHelper(result).exists():
            DirectoryStructureRestoration(self.parent_dirname).start()
        return result

    @ensure_status_is_given
    def generate_hosts_file(self) -> "StatusFileGenerator":
        """
        Generates the hosts file.
        """

        our_dataset = self.status.to_dict()
        our_dataset["ip"] = self.hosts_ip

        if not hasattr(self.status, "ip_syntax") or not self.status.ip_syntax:
            location = os.path.join(
                self.get_output_basedir(),
                PyFunceble.cli.storage.OUTPUTS.hosts.directory,
                self.status.status.upper(),
                PyFunceble.cli.storage.OUTPUTS.hosts.filename,
            )
        else:
            location = os.path.join(
                self.get_output_basedir(),
                PyFunceble.cli.storage.OUTPUTS.hosts.directory,
                self.status.status.upper(),
                PyFunceble.cli.storage.OUTPUTS.hosts.ip_filename,
            )

        self.file_printer.destination = location
        self.file_printer.dataset = our_dataset
        self.file_printer.template_to_use = "hosts"
        self.file_printer.print_interpolated_line()

        return self

    @ensure_status_is_given
    def generate_plain_file(self) -> "StatusFileGenerator":
        """
        Generates the plain file.
        """

        if not hasattr(self.status, "ip_syntax") or not self.status.ip_syntax:
            location = os.path.join(
                self.get_output_basedir(),
                PyFunceble.cli.storage.OUTPUTS.domains.directory,
                self.status.status.upper(),
                PyFunceble.cli.storage.OUTPUTS.domains.filename,
            )

            self.file_printer.destination = location
            self.file_printer.dataset = self.status.to_dict()
            self.file_printer.template_to_use = "plain"
            self.file_printer.print_interpolated_line()

        return self

    @ensure_status_is_given
    def generate_analytic_file(self) -> "StatusFileGenerator":
        """
        Generates the analytic files.
        """

        locations_data_and_template = []

        # pylint: disable=line-too-long

        if hasattr(self.status, "http_status_code") and self.status.http_status_code:
            if PyFunceble.facility.ConfigLoader.is_already_loaded():
                http_code_dataset = PyFunceble.storage.HTTP_CODES
            else:
                http_code_dataset = PyFunceble.storage.STD_HTTP_CODES

            if (
                self.status.http_status_code in http_code_dataset.list.potentially_down
                or self.status.status
                in (PyFunceble.storage.STATUS.down, PyFunceble.storage.STATUS.invalid)
            ):
                locations_data_and_template.append(
                    (
                        os.path.join(
                            self.get_output_basedir(),
                            PyFunceble.cli.storage.OUTPUTS.analytic.directories.parent,
                            PyFunceble.cli.storage.OUTPUTS.analytic.directories.potentially_down,
                            PyFunceble.cli.storage.OUTPUTS.analytic.filenames.potentially_down,
                        ),
                        "plain",
                        self.status.to_dict(),
                    )
                )

            if self.status.http_status_code in http_code_dataset.list.up:
                locations_data_and_template.append(
                    (
                        os.path.join(
                            self.get_output_basedir(),
                            PyFunceble.cli.storage.OUTPUTS.analytic.directories.parent,
                            PyFunceble.cli.storage.OUTPUTS.analytic.directories.potentially_down,
                            PyFunceble.cli.storage.OUTPUTS.analytic.filenames.up,
                        ),
                        "plain",
                        self.status.to_dict(),
                    )
                )

            if (
                self.status.http_status_code in http_code_dataset.list.potentially_up
                or self.status.status == PyFunceble.storage.STATUS.down
            ):
                locations_data_and_template.append(
                    (
                        os.path.join(
                            self.get_output_basedir(),
                            PyFunceble.cli.storage.OUTPUTS.analytic.directories.parent,
                            PyFunceble.cli.storage.OUTPUTS.analytic.directories.potentially_down,
                            PyFunceble.cli.storage.OUTPUTS.analytic.filenames.potentially_up,
                        ),
                        "plain",
                        self.status.to_dict(),
                    )
                )

        if self.test_dataset and "from_inactive" in self.test_dataset:
            # Let's generate the supicious file :-)
            if self.status.status in [
                PyFunceble.storage.STATUS.up,
                PyFunceble.storage.STATUS.valid,
                PyFunceble.storage.STATUS.sane,
            ]:
                locations_data_and_template.append(
                    (
                        os.path.join(
                            self.get_output_basedir(),
                            PyFunceble.cli.storage.OUTPUTS.analytic.directories.parent,
                            PyFunceble.cli.storage.OUTPUTS.analytic.directories.suspicious,
                            PyFunceble.cli.storage.OUTPUTS.analytic.filenames.suspicious,
                        ),
                        "plain",
                        self.status.to_dict(),
                    )
                )

        for file_location, template, our_dataset in locations_data_and_template:
            self.file_printer.destination = file_location
            self.file_printer.dataset = our_dataset
            self.file_printer.template_to_use = template
            self.file_printer.print_interpolated_line()

        return self

    @ensure_status_is_given
    def generate_splitted_status_file(self) -> "StatusFileGenerator":
        """
        Generates the splitted status file.
        """

        self.file_printer.destination = os.path.join(
            self.get_output_basedir(),
            PyFunceble.cli.storage.OUTPUTS.splitted.directory,
            self.status.status.upper(),
        )

        if not PlatformUtility.is_unix():
            self.file_printer.destination += ".txt"

        self.file_printer.template_to_use = "all"
        self.file_printer.dataset = self.status.to_dict()
        self.file_printer.print_interpolated_line()

        return self

    @ensure_status_is_given
    def generate_unified_status_file(self) -> "StatusFileGenerator":
        """
        Generates the unified status file.
        """

        self.file_printer.destination = os.path.join(
            self.get_output_basedir(),
            PyFunceble.cli.storage.RESULTS_RAW_FILE,
        )
        self.file_printer.template_to_use = "all"
        self.file_printer.dataset = self.status.to_dict()
        self.file_printer.print_interpolated_line()

        return self

    def start(self) -> "StatusFileGenerator":
        """
        Starts the generation of everything possible.
        """

        if self.allow_hosts_files:
            self.generate_hosts_file()

        if self.allow_plain_files:
            self.generate_plain_file()

        if self.allow_analytic_files:
            self.generate_analytic_file()

        if self.allow_unified_file:
            self.generate_unified_status_file()
        else:
            self.generate_splitted_status_file()



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides some utilities related to the CI.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import PyFunceble.facility
from PyFunceble.cli.continuous_integration.base import ContinuousIntegrationBase
from PyFunceble.cli.continuous_integration.github_actions import GitHubActions
from PyFunceble.cli.continuous_integration.gitlab_ci import GitLabCI
from PyFunceble.cli.continuous_integration.jenkins import Jenkins
from PyFunceble.cli.continuous_integration.travis_ci import TravisCI


def ci_object(*args, **kwargs) -> ContinuousIntegrationBase:
    """
    A placeholder which provides the CI object to use.
    """

    known_objects = [Jenkins, GitHubActions, TravisCI, GitLabCI]
    result = None

    for known in known_objects:
        result = known(*args, **kwargs)
        result.guess_all_settings()

        PyFunceble.facility.Logger.debug("Checking if %r is authorized.", result)

        if result.is_authorized():
            PyFunceble.facility.Logger.debug(
                "%r is authorized. Using it as CI object.", result
            )
            return result

    PyFunceble.facility.Logger.debug(
        "No known CI object authorized. Using: %r", known_objects[0]
    )

    return known_objects[0](*args, **kwargs)



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the download helpers.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from typing import Optional

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

import PyFunceble.helpers.exceptions
from PyFunceble.helpers.file import FileHelper


class DownloadHelper:
    """
    Simplification of the downloads.

    :param str url:
    :param int retry:
        The number of time we have to retry before raising an exception.
    """

    _url: Optional[str] = None
    _certificate_validation: bool = True
    _retries: int = 3

    def __init__(
        self,
        url: Optional[str] = None,
        *,
        certificate_validation: bool = True,
        retries: int = 3,
    ) -> None:
        if url:
            self.url = url

        if certificate_validation:
            self.certificate_validation = certificate_validation

        if retries:
            self.retries = retries

    @property
    def url(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_url` attribute.
        """

        return self._url

    @url.setter
    def url(self, value: str) -> None:
        """
        Sets the url to work with.

        :param value:
            The URL to set.

        :raise TypeError:
            When :code:`value` is not a :py:class:`str`.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        self._url = value

    def set_url(self, value: str) -> "DownloadHelper":
        """
        Sets the url to work with.

        :param value:
            The URL to set.
        """

        self.url = value

        return self

    @property
    def certificate_validation(self) -> bool:
        """
        Provides the current state of the :code:`certificate_validation`
        attribute.
        """

        return self._certificate_validation

    @certificate_validation.setter
    def certificate_validation(self, value: bool) -> None:
        """
        Sets the value of the certificate validation.

        :param value:
            The value to set.

        :raise TypeError:
            When :code:`value` is not a :py:class:`bool`
        """

        if not isinstance(value, bool):
            raise TypeError(f"<value> should be {bool}, {type(value)} given.")

        self._certificate_validation = value

    def set_certificate_validation(self, value: bool) -> "DownloadHelper":
        """
        Sets the value of the certificate validation.

        :param value:
            The value to set.
        """

        self.certificate_validation = value

        return self

    @property
    def retries(self) -> int:
        """
        Provides the current state of the :code:`_retries` attributes.
        """

        return self._retries

    @retries.setter
    def retries(self, value: int) -> None:
        """
        Sets the number of retries we are allowed to perform before raising an
        exception.

        :param value:
            The number of retry to apply.i

        :raise TypeError:
            When :code:`value` is not a :py:class:`int`.
        :raise ValueError:
            When :code:`value` is lower than :code:`0`.
        """

        if not isinstance(value, int):
            raise TypeError(f"<value> should be {int}, {type(value)} given.")

        if value <= 0:
            raise ValueError("<value> should greater than zero.")

        self._retries = value

    def set_retries(self, value: int) -> "DownloadHelper":
        """
        Sets the number of retries we are allowed to perform before raising an
        exception.

        :param value:
            The number of retry to apply.i
        """

        self.retries = value

        return self

    def download_text(self, *, destination: Optional[str] = None) -> str:
        """
        Download the body of the set url.

        .. note::
            if :code:`destination` is set to :code:`None`,
            we only return the output.

            Otherwise, we save the output into the given
            destination, but we also return the output.

        :param destination: The download destination.

        :raise UnableToDownload: When could not unable to download the URL.
        """

        session = requests.Session()

        retries = Retry(total=self.retries, backoff_factor=3)
        adapter = HTTPAdapter(max_retries=retries)

        session.mount("http://", adapter)
        session.mount("https://", adapter)

        req = session.get(self.url, verify=self.certificate_validation)

        if req.status_code == 200:
            response = req.text

            if destination and isinstance(destination, str):
                FileHelper(destination).write(req.text, overwrite=True)

            adapter.close()
            req.close()
            return response

        adapter.close()
        session.close()
        raise PyFunceble.helpers.exceptions.UnableToDownload(
            f"{req.url} (retries: {self.retries} | status code: {req.status_code})"
        )



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides our interface for quering the WHOIS Record of a given subject.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import functools
import socket
from typing import Optional, Union

from PyFunceble.dataset.iana import IanaDataset
from PyFunceble.query.record.whois import WhoisQueryToolRecord
from PyFunceble.query.whois.converter.expiration_date import ExpirationDateExtractor
from PyFunceble.query.whois.converter.registrar import RegistarExtractor

# pylint: disable=protected-access


class WhoisQueryTool:
    """
    Provides the interface to get the WHOIS record of a given subject.
    """

    BUFFER_SIZE: int = 4096
    STD_PORT: int = 43

    expiration_date_extractor: Optional[ExpirationDateExtractor] = None
    registrar_extractor: Optional[RegistarExtractor] = None
    iana_dataset: Optional[IanaDataset] = None

    _subject: Optional[str] = None
    _server: Optional[str] = None
    _query_timeout: float = 5.0
    _expiration_date: Optional[str] = None
    _record: Optional[str] = None
    _registrar: Optional[str] = None

    lookup_record: Optional[WhoisQueryToolRecord] = None

    def __init__(
        self,
        subject: Optional[str] = None,
        *,
        server: Optional[str] = None,
        query_timeout: Optional[float] = None,
    ) -> None:
        self.registrar_extractor = RegistarExtractor()
        self.expiration_date_extractor = ExpirationDateExtractor()
        self.iana_dataset = IanaDataset()

        if subject is not None:
            self.subject = subject

        if server is not None:
            self.server = server

        if query_timeout is not None:
            self.set_query_timeout(query_timeout)

    def ensure_subject_is_given(func):  # pylint: disable=no-self-argument
        """
        Ensures that the subject is given before running the decorated method.

        :raise TypeError:
            If the subject is not a string.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):  # pragma: no cover ## Just safety.
            if not isinstance(self.subject, str):
                raise TypeError(
                    f"<self.subject> should be {str}, {type(self.subject)} given."
                )

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    def update_lookup_record(func):  # pylint: disable=no-self-argument
        """
        Ensures that a clean record is generated after the execution of
        the decorated method.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            result = func(self, *args, **kwargs)  # pylint: disable=not-callable

            if self.lookup_record is None or self.subject != self.lookup_record.subject:
                self.lookup_record = WhoisQueryToolRecord(port=self.STD_PORT)
                self.lookup_record.subject = self.subject

            if self.query_timeout != self.lookup_record.query_timeout:
                self.lookup_record.query_timeout = self.query_timeout

            if self.server != self.lookup_record.server:
                self.lookup_record.server = self.server

            if self._expiration_date != self.lookup_record.expiration_date:
                self.lookup_record.expiration_date = self._expiration_date

            if self._registrar != self.lookup_record.registrar:
                self.lookup_record.registrar = self._registrar

            return result

        return wrapper

    def reset_record(func):  # pylint: disable=no-self-argument
        """
        Resets the record before executing the decorated method.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if self.lookup_record:
                self.lookup_record.record = self._record = None
                self.lookup_record.expiration_date = self._expiration_date = None
                self.lookup_record.registrar = self._registrar = None

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    def query_record(func):  # pylint: disable=no-self-argument
        """
        Queries the record before executing the decorated method.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):  # pragma: no cover ## It just works.
            if not self.lookup_record or self.lookup_record.record is None:
                self.query()

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    @property
    def subject(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_subject` attribute.
        """

        return self._subject

    @subject.setter
    @reset_record
    @update_lookup_record
    def subject(self, value: str) -> None:
        """
        Sets the subject to work with.

        :param value:
            The subject to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise ValueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empty.")

        self._subject = value

    def set_subject(self, value: str) -> "WhoisQueryTool":
        """
        Sets the subject to work with.

        :param value:
            The subject to set.
        """

        self.subject = value

        return self

    @property
    def server(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_server` attribute.
        """

        return self._server

    @server.setter
    @reset_record
    @update_lookup_record
    def server(self, value: str) -> None:
        """
        Sets the server to communicate with.

        :param value:
            The server to work with.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise ValueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empty.")

        self._server = value

    def set_server(self, value: str) -> "WhoisQueryTool":
        """
        Sets the server to communicate with.

        :param value:
            The server to work with.
        """

        self.server = value

        return self

    @property
    def query_timeout(self) -> float:
        """
        Provides the current state of the :code:`_query_timeout` attribute.
        """

        return self._query_timeout

    @query_timeout.setter
    def query_timeout(self, value: Union[float, int]) -> None:
        """
        Sets the query_timeout to apply.

        :param value:
            The query_timeout to apply.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`int`
            nor :py:class:`float`.
        :raise ValueError:
            When the given :code:`value` is less than `1`.
        """

        if not isinstance(value, (int, float)):
            raise TypeError(f"<value> should be {int} or {float}, {type(value)} given.")

        if value < 1:
            raise ValueError(f"<value> ({value!r}) should be less than 1.")

        self._query_timeout = float(value)

    def set_query_timeout(self, value: Union[float, int]) -> "WhoisQueryTool":
        """
        Sets the query_timeout to apply.

        :param value:
            The query_timeout to apply.
        """

        self.query_timeout = value

        return self

    @property
    @query_record
    @update_lookup_record
    def expiration_date(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_expiration_date` attribute.
        """

        if self.lookup_record.record:
            self._expiration_date = self.expiration_date_extractor.set_data_to_convert(
                self.lookup_record.record
            ).get_converted()

        return self._expiration_date

    @property
    @query_record
    def record(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_record` attribute.
        """

        return self.lookup_record.record

    @property
    @query_record
    @update_lookup_record
    def registrar(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_registrar` attribute.
        """

        if self.lookup_record.record:
            self._registrar = self.registrar_extractor.set_data_to_convert(
                self.lookup_record.record
            ).get_converted()

        return self._registrar

    @ensure_subject_is_given
    def get_whois_server(
        self,
    ) -> Optional[str]:  # pragma: no cover ## Test the underlying method instead.
        """
        Provides the whois server to work with.
        """

        if self.subject.endswith("."):
            extension = self.subject[:-1]
        else:
            extension = self.subject

        extension = extension[extension.rfind(".") + 1 :]

        return self.iana_dataset.get_whois_server(extension)

    def get_lookup_record(
        self,
    ) -> Optional[WhoisQueryToolRecord]:
        """
        Provides the current query record.
        """

        return self.lookup_record

    @ensure_subject_is_given
    @update_lookup_record
    def query(
        self,
    ) -> str:  # pragma: no cover ## The effect of the response of this method
        ## are more important.
        """
        Queries the WHOIS record and return the current object.
        """

        if self.lookup_record.record is None:

            if not self.server:
                whois_server = self.get_whois_server()
            else:
                whois_server = self.server

            if whois_server:
                self.lookup_record.server = whois_server
                self.lookup_record.query_timeout = self.query_timeout

                req = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                req.settimeout(self.query_timeout)

                try:
                    req.connect((whois_server, self.STD_PORT))
                    req.send(f"{self.subject}\r\n".encode())

                    response = "".encode()

                    while True:
                        try:
                            data = req.recv(self.BUFFER_SIZE)
                        except (ConnectionResetError, socket.timeout):
                            break

                        response += data

                        if not data:
                            break

                    req.close()

                    try:
                        self.lookup_record.record = self._record = response.decode()
                    except UnicodeDecodeError:
                        # Note: Because we don't want to deal with other issue, we
                        # decided to use `replace` in order to automatically replace
                        # all non utf-8 encoded characters.
                        self.lookup_record.record = self._record = response.decode(
                            "utf-8", "replace"
                        )
                except socket.error:
                    pass

            if self.lookup_record.record is None or not self.lookup_record.record:
                self.lookup_record.record = self._record = ""
                self.lookup_record.expiration_date = self._expiration_date = ""
                self.lookup_record.registrar = self._record = ""
            else:
                self.lookup_record.expiration_date = self.expiration_date
                self.lookup_record.registrar = self.registrar

        return self.lookup_record.record

    @query_record
    def get_record(
        self,
    ) -> Optional[str]:  # pragma: no cover ## Just a dataclass attribute.
        """
        Provides the current record.

        Side Effect:
            The record will be queried if it is non existent.
        """

        return self.lookup_record.record



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the regular expressions helpers.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import re
from typing import List, Optional, Union


class RegexHelper:
    """
    Simplify the regex matching and usage.

    :param str regex: The regex to use.
    :param escape_regex: Escapes the given regex.
    """

    _regex: Optional[str] = None
    escape_regex: bool = False

    def __init__(self, regex: Optional[str] = None, escape_regex: bool = False):
        self.escape_regex = escape_regex

        if regex is not None:
            self.regex = regex

    @property
    def regex(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_regex` attribute.
        """

        return self._regex

    @regex.setter
    def regex(self, value: str) -> None:
        """
        Sets the regex to work with.

        :param value:
            The regex to work with.

        :raise TypeError:
            When :code:`value` is not :py:class:`str`.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(str)} given.")

        if not self.escape_regex:
            self._regex = value
        else:
            self._regex = re.escape(value)

    def set_regex(self, value: str) -> "RegexHelper":
        """
        Sets the regex to work with.

        :param value:
            The regex to work with.

        :raise TypeError:
            When :code:`value` is not :py:class:`str`.
        """

        self.regex = value

        return self

    def get_not_matching_list(self, data: List[str]) -> List[str]:
        """
        Returns the strings which does not the match the regex
        in the given data.
        """

        pre_result = re.compile(self.regex)

        return [x for x in data if not pre_result.search(str(x))]

    def get_matching_list(self, data: List[str]) -> List[str]:
        """
        Returns the strings which does the match the regex
        in the given data.
        """

        pre_result = re.compile(self.regex)

        return [x for x in data if pre_result.search(str(x))]

    def match(
        self,
        data: str,
        *,
        rematch: bool = False,
        group: int = 0,
        return_match: bool = True,
    ) -> Union[bool, str, List[str]]:
        """
        Checks if the given data match the given regex string.

        :param data: The data to work with.
        :param rematch:
            The equivalent of the $BASH_REMATCH but in Python.

            It's basically a list of all groups.
        :param group:
            The group to return when return_match is set to :code:`True`.
        :param return_match:
            Return the part that match the given regex string.
        """

        result = []
        to_match = re.compile(self.regex)

        if rematch:
            pre_result = to_match.findall(data)
        else:
            pre_result = to_match.search(data)

        if return_match and pre_result:
            if rematch:
                for res in pre_result:
                    if isinstance(res, tuple):
                        result.extend(list(res))
                    else:  # pragma: no cover ## Safety
                        result.append(res)

                if group != 0:
                    return result[group]
            else:
                result = pre_result.group(group).strip()

            return result

        if not return_match and pre_result:
            return True
        return False

    def replace_match(
        self,
        data: str,
        replacement: str,
        *,
        occurences: int = 0,
        multiline: bool = False,
    ) -> str:
        """
        Replaces the string which match the regex string with
        the given replacement.

        :param data: The data to work with.
        :param replacement: The replacement of the matched regex.
        :param occurences:
            The number of occurences to replace.

            .. note::
                :code:`0` means all occurences.
        """

        if isinstance(replacement, str):
            return re.sub(
                self.regex,
                replacement,
                data,
                occurences,
                flags=re.MULTILINE if multiline else 0,
            )
        return data

    def split(self, data: str) -> List[str]:
        """
        Split the reference of the given regex.

        :param str data: The data to work with.
        :rtype: list
        """

        return re.split(self.regex, data)



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the subdomain syntax checker.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from PyFunceble.checker.syntax.domain_base import DomainSyntaxCheckerBase
from PyFunceble.helpers.regex import RegexHelper


class SubDomainSyntaxChecker(DomainSyntaxCheckerBase):
    """
    Provides an interface to check the syntax of a subdomain.

    :param str subject:
        Optional, The subject to work with.
    """

    # pylint: disable=line-too-long
    REGEX_VALID_SUBDOMAIN: str = r"^(?=.{0,253}$)(([a-z0-9_][a-z0-9-_]{0,61}[a-z0-9_-]|[a-z0-9])\.)+((?=.*)([a-z0-9][a-z0-9-]{0,61}[a-z0-9](?:\.)?|[a-z0-9](?:\.)?))$"

    @DomainSyntaxCheckerBase.ensure_subject_is_given
    def is_valid(self) -> bool:
        """
        Validate the given subject if exists.
        """

        extension = self.get_extension()

        if (
            not extension
            or extension not in self.iana_dataset
            and extension not in self.SPECIAL_USE_DOMAIN_NAMES_EXTENSIONS
        ):
            return False

        subject_without_extension = self.idna_subject[: self.last_point_index]
        subject_without_suffix, suffix = self.get_subject_without_suffix(
            self.idna_subject, extension
        )

        if subject_without_suffix:
            if suffix.count(".") >= 2:
                return RegexHelper(self.REGEX_VALID_SUBDOMAIN).match(
                    subject_without_extension, return_match=False
                )

            if "." in subject_without_suffix:
                return RegexHelper(self.REGEX_VALID_SUBDOMAIN).match(
                    self.idna_subject, return_match=False
                )

            return False

        if "." in subject_without_extension:
            return RegexHelper(self.REGEX_VALID_SUBDOMAIN).match(
                subject_without_extension, return_match=False
            )

        return False



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides our registrar extrator.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from typing import Any, List, Optional

from PyFunceble.helpers.regex import RegexHelper
from PyFunceble.query.whois.converter.base import ConverterBase


class RegistarExtractor(ConverterBase):
    """
    Provides an interface for the extration of the registrar.
    """

    PATTERNS: List[str] = [
        r"authorized\s+agency(\s+|):(.*)",
        r"domain\s+support(\s+|):(.*)",
        r"registrar\s+name(\s+|):(.*)",
        r"registrar_name(\s+|):(.*)",
        r"registrar(\s+|):(.*)",
        r"registrar\.+(\s+|):(.*)",
        r"registration\s+service\s+provider(\s+|):(.*)",
        r"sponsoring\s+registrar(\s+|):(.*)",
        r"sponsoring\s+registrar\s+organization(\s+|):(.*)",
    ]

    @ConverterBase.data_to_convert.setter
    def data_to_convert(self, value: Any) -> None:
        """
        Sets the data to convert and work with.

        :param value:
            The record to work with.

        :raise TypeError:
            When the given data to convert is not :py:class:`str`.
        :raise ValueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empty.")

        # pylint: disable=no-member
        super(RegistarExtractor, self.__class__).data_to_convert.fset(self, value)

    def __get_line(self) -> Optional[str]:
        """
        Tries to get the registrar line from the given record.
        """

        for regex in self.PATTERNS:
            registrar_line = RegexHelper(r"(?i)" + regex).match(
                self.data_to_convert, return_match=True, rematch=True, group=0
            )

            if not registrar_line:
                continue

            return registrar_line
        return None

    @ConverterBase.ensure_data_to_convert_is_given
    def get_converted(self) -> Optional[str]:
        """
        Provides the registrar of the record (if found).
        """

        registrar_line = self.__get_line()

        if registrar_line:
            try:
                return [x.strip() for x in registrar_line if x.strip()][0]
            except IndexError:
                pass

        return None



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides our producer worker. This is the description of a single producer worker.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import datetime
import os
from typing import Any, Optional, Tuple

from domain2idna import domain2idna

import PyFunceble.facility
import PyFunceble.storage
from PyFunceble.checker.status_base import CheckerStatusBase
from PyFunceble.checker.utils.whois import get_whois_dataset_object
from PyFunceble.cli.filesystem.counter import FilesystemCounter
from PyFunceble.cli.filesystem.printer.file import FilePrinter
from PyFunceble.cli.filesystem.printer.stdout import StdoutPrinter
from PyFunceble.cli.filesystem.registrar_counter import RegistrarCounter
from PyFunceble.cli.filesystem.status_file import StatusFileGenerator
from PyFunceble.cli.processes.workers.base import WorkerBase
from PyFunceble.cli.utils.stdout import get_template_to_use, print_single_line
from PyFunceble.cli.utils.testing import (
    get_continue_databaset_object,
    get_inactive_dataset_object,
)
from PyFunceble.dataset.autocontinue.base import ContinueDatasetBase
from PyFunceble.dataset.autocontinue.csv import CSVContinueDataset
from PyFunceble.dataset.inactive.base import InactiveDatasetBase
from PyFunceble.dataset.whois.base import WhoisDatasetBase
from PyFunceble.query.collection import CollectionQueryTool


class ProducerWorker(WorkerBase):
    """
    Provides our producer worker. The objective of this worker is to provides
    a single worker (or process if you prefer) which will be used to handle
    the production of output to stdout or files.
    """

    STD_NAME: str = "pyfunceble_producer_worker"

    stdout_printer: Optional[StdoutPrinter] = None
    file_printer: Optional[FilePrinter] = None
    whois_dataset: Optional[WhoisDatasetBase] = None
    inactive_dataset: Optional[InactiveDatasetBase] = None
    continue_dataset: Optional[ContinueDatasetBase] = None
    status_file_generator: Optional[StatusFileGenerator] = None
    counter: Optional[FilesystemCounter] = None
    registrar_counter: Optional[RegistrarCounter] = None
    collection_query_tool: Optional[CollectionQueryTool] = None

    header_already_printed: Optional[bool] = None

    INACTIVE_STATUSES: Tuple[str, ...] = (
        PyFunceble.storage.STATUS.down,
        PyFunceble.storage.STATUS.invalid,
    )

    def __post_init__(self) -> None:
        self.stdout_printer = StdoutPrinter()
        self.file_printer = FilePrinter()
        self.whois_dataset = get_whois_dataset_object(db_session=self.db_session)
        self.inactive_dataset = get_inactive_dataset_object(db_session=self.db_session)
        self.continue_dataset = get_continue_databaset_object(
            db_session=self.db_session
        )
        self.status_file_generator = StatusFileGenerator().guess_all_settings()
        self.counter = FilesystemCounter()
        self.registrar_counter = RegistrarCounter()
        self.collection_query_tool = CollectionQueryTool()

        self.header_already_printed = False

        return super().__post_init__()

    @staticmethod
    def should_we_ignore(test_result: CheckerStatusBase) -> bool:
        """
        Checks if we should ignore the given datasets.

        :param test_result:
            The test result to check.
        """

        return isinstance(test_result, str) and test_result.startswith("ignored_")

    @staticmethod
    def should_we_print_status_to_stdout(status: str) -> bool:
        """
        Checks if we are allows to print the given status (to stdout).

        :param status:
            The status to check.
        """

        if isinstance(
            PyFunceble.storage.CONFIGURATION.cli_testing.display_mode.status, list
        ):
            to_keep = PyFunceble.storage.CONFIGURATION.cli_testing.display_mode.status
        else:
            to_keep = [PyFunceble.storage.CONFIGURATION.cli_testing.display_mode.status]

        to_keep = [x.upper() for x in to_keep]
        status = status.upper()

        return "ALL" in to_keep or status in to_keep

    def should_we_block_status_file_printer(
        self, test_dataset: dict, test_result: CheckerStatusBase
    ) -> bool:
        """
        Checks if we should block the file printer.

        The reason behindn this is that we don't want to generate an output
        when a subject was already into the inactive database.
        """

        return (
            "from_inactive" in test_dataset
            and test_result.status in self.INACTIVE_STATUSES
        )

    def run_stdout_printer(self, test_result: CheckerStatusBase) -> None:
        """
        Runs the stdout printer (if necessary).

        :param test_result:
            The rest result dataset.
        """

        if not PyFunceble.storage.CONFIGURATION.cli_testing.display_mode.quiet:
            # pylint: disable=line-too-long

            if self.should_we_print_status_to_stdout(test_result.status):
                self.stdout_printer.template_to_use = get_template_to_use()

                if not self.header_already_printed:
                    self.stdout_printer.print_header()
                    self.header_already_printed = True

                self.stdout_printer.set_dataset(
                    test_result.to_dict()
                ).print_interpolated_line()
            else:
                print_single_line()
        else:
            print_single_line()

    def run_whois_backup(self, test_result: CheckerStatusBase) -> None:
        """
        Runs the backup or update of the WHOIS record in our dataset storage.

        :param test_result:
            The test result.
        """

        if (
            hasattr(test_result, "expiration_date")
            and test_result.expiration_date
            and test_result.whois_record
        ):
            # Note: The whois record is always given if the status does not come
            # from the database.

            self.whois_dataset.update(
                {
                    "subject": test_result.subject,
                    "idna_subject": test_result.idna_subject,
                    "expiration_date": test_result.expiration_date,
                    "epoch": str(
                        datetime.datetime.strptime(
                            test_result.expiration_date, "%d-%b-%Y"
                        ).timestamp()
                    ),
                    "registrar": test_result.registrar,
                }
            )

    def run_inactive_backup(
        self, test_dataset: dict, test_result: CheckerStatusBase
    ) -> None:
        """
        Runs the backup or update of the Inactive dataset storage.

        The idea is that if the status is OK (active), we just remove it from
        the dataset storage. Otherwise, we just keep it in there :-)
        """

        if test_dataset["type"] != "single" and self.inactive_dataset.authorized:
            dataset = test_result.to_dict()
            dataset.update(test_dataset)

            PyFunceble.facility.Logger.debug("Test Dataset: %r", test_dataset)
            PyFunceble.facility.Logger.debug("Test Result: %r", test_result)

            if "from_inactive" in dataset:
                if test_result.status in self.INACTIVE_STATUSES:
                    if dataset["destination"]:
                        # Note: This handles the case that someone try to test a
                        # single subject.
                        # In fact, we should not generate any file if the output
                        # directory is not given.
                        self.inactive_dataset.update(dataset)
                else:
                    self.inactive_dataset.remove(dataset)
            elif (
                test_result.status in self.INACTIVE_STATUSES and dataset["destination"]
            ):
                self.inactive_dataset.update(dataset)

    def run_continue_backup(
        self, test_dataset: dict, test_result: CheckerStatusBase
    ) -> None:
        """
        Runs the backup or update of the auto-continue dataset storage.
        """

        if test_dataset["type"] != "single" and self.continue_dataset.authorized:
            dataset = test_result.to_dict()
            dataset.update(test_dataset)

            if (
                isinstance(self.continue_dataset, CSVContinueDataset)
                and dataset["output_dir"]
            ):
                # Note: This handles the case that someone try to test a single subject.
                # In fact, we should not generate any file if the output directory
                # is not given.
                #
                # The exception handler is just there to catch the case that the
                # method is not implemented because we are more likely using an
                # alternative backup (for example mariadb).

                self.continue_dataset.set_base_directory(dataset["output_dir"])
            self.continue_dataset.update(dataset)

    def run_ignored_file_printer(self, test_dataset: dict, test_result: str) -> None:
        """
        Runs the analytic behind the file printer.

        .. warning::
            Thie method assume that the givne dataset is ignored from the normal
            file printer.
        """

        if (
            test_result == "ignored_inactive"
            and test_dataset["destination"]
            and not PyFunceble.storage.CONFIGURATION.cli_testing.file_generation.no_file
        ):
            if "idna_subject" not in test_dataset:
                # The printer always prints the idna subject.
                test_dataset["idna_subject"] = domain2idna(test_dataset["subject"])

            # We use this object just to get the output directory :-)
            self.status_file_generator.set_parent_dirname(test_dataset["destination"])

            # We now generate the file which let the end-user know that
            # we ignored the test of the current test dataset because there
            # is no need to retest it yet.
            # pylint: disable=line-too-long
            self.file_printer.destination = os.path.join(
                self.status_file_generator.get_output_basedir(),
                PyFunceble.cli.storage.OUTPUTS.logs.directories.parent,
                PyFunceble.cli.storage.OUTPUTS.logs.filenames.inactive_not_retested,
            )

            self.file_printer.template_to_use = "plain"
            self.file_printer.dataset = test_dataset
            self.file_printer.print_interpolated_line()

    def run_status_file_printer(
        self, test_dataset: dict, test_result: CheckerStatusBase
    ) -> None:
        """
        Runs the status file printer.
        """

        if (
            test_dataset["destination"]
            and not PyFunceble.storage.CONFIGURATION.cli_testing.file_generation.no_file
        ):
            previous_allow_hosts_file = self.status_file_generator.allow_hosts_files
            previous_allow_plain_file = self.status_file_generator.allow_plain_files

            self.status_file_generator.parent_dirname = test_dataset["destination"]
            self.status_file_generator.status = test_result
            self.status_file_generator.test_dataset = test_dataset

            if test_dataset["subject_type"] == "url":
                self.status_file_generator.allow_hosts_files = False
                self.status_file_generator.allow_plain_files = True

            self.status_file_generator.start()
            self.status_file_generator.allow_hosts_files = previous_allow_hosts_file
            self.status_file_generator.allow_plain_files = previous_allow_plain_file

    def run_counter(self, test_dataset: dict, test_result: CheckerStatusBase) -> None:
        """
        Runs the counter of the current file.
        """

        if (
            test_dataset["destination"]
            and not PyFunceble.storage.CONFIGURATION.cli_testing.file_generation.no_file
        ):
            # Note: We don't want hidden data to be counted.

            self.counter.set_parent_dirname(test_dataset["destination"]).count(
                test_result
            )

            if hasattr(test_result, "registrar") and test_result.registrar:
                self.registrar_counter.set_parent_dirname(
                    test_dataset["destination"]
                ).count(test_result.registrar)

    def target(self, consumed: Any) -> Optional[Tuple[Any, ...]]:
        if not isinstance(consumed, tuple):
            PyFunceble.facility.Logger.info(
                "Skipping latest dataset because consumed data was not a tuple."
            )
            return None

        # Just for human brain.
        test_dataset, test_result = consumed

        if not isinstance(test_dataset, dict):
            PyFunceble.facility.Logger.info(
                "Skipping because test dataset is not a dictionnary."
            )
            return None

        if self.should_we_ignore(test_result):
            PyFunceble.facility.Logger.info(
                "Ignored test dataset. Reason: No output wanted."
            )
            self.run_ignored_file_printer(test_dataset, test_result)
            return None

        if not isinstance(test_result, CheckerStatusBase):
            PyFunceble.facility.Logger.info(
                "Skipping latest dataset because consumed status is not "
                "a status object.."
            )
            return None

        if (
            PyFunceble.storage.CONFIGURATION.collection.push
            and test_result.status_source != "COLLECTION"
        ):
            self.collection_query_tool.push(test_result)

        if not PyFunceble.storage.CONFIGURATION.cli_testing.chancy_tester:
            self.run_whois_backup(test_result)
            self.run_inactive_backup(test_dataset, test_result)
            self.run_continue_backup(test_dataset, test_result)

        if not self.should_we_block_status_file_printer(test_dataset, test_result):
            self.run_status_file_printer(test_dataset, test_result)

            if not PyFunceble.storage.CONFIGURATION.cli_testing.chancy_tester:
                self.run_counter(test_dataset, test_result)

        self.run_stdout_printer(test_result)
        test_dataset["from_producer"] = True

        return test_dataset, test_result



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides our expiration date extracter.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from typing import List, Optional, Tuple

from PyFunceble.helpers.regex import RegexHelper
from PyFunceble.query.whois.converter.base import ConverterBase
from PyFunceble.query.whois.converter.digit2digits import Digit2Digits
from PyFunceble.query.whois.converter.month2unified import Month2Unified


class ExpirationDateExtractor(ConverterBase):
    """
    Provides an interface for the extraction of the expiration date.
    """

    PATTERNS: List[str] = [
        r"\[expires\s+on\](.*)",
        r"data\s+de\s+expiração\s+\/\s+Expiration\s+Date\s+\(dd\/mm\/yyyy\):(.*)",
        r"date\s+d'expiration:(.*)",
        r"domain_datebilleduntil:(.*)",
        r"Exp\s+date:(.*)",
        r"expiration:\.+(.*)",
        r"expiration\s+date:(.*)",
        r"expiration\s+date\s+:(.*)",
        r"expiration\s+time:(.*)",
        r"expire-date:(.*)",
        r"expire:(.*)",
        r"expire\s+date:(.*)",
        r"expire\s+on:(.*)",
        r"expired:(.*)",
        r"expires:(.*)",
        r"expires\.+:(.*)",
        r"expires\s+at:(.*)",
        r"expires\s+on:(.*)",
        r"expires\s+on\.+:(.*)",
        r"expiry\s+:(.*)",
        r"expiry\s+date:(.*)",
        r"fecha\s+de\s+expiración\s+\(expiration\s+date\):(.*)",
        r"fecha\s+de\s+vencimiento:(.*)",
        r"free-date(.*)",
        r"record\s+expires\s+on(.*)(\(YYYY-MM-DD\))",
        r"registry\s+expiry\s+date:(.*)",
        r"renewal:(.*)",
        r"renewal\s+date:(.*)",
        r"status:\s+OK-UNTIL(.*)",
        r"valid-date(.*)",
        r"validity:(.*)",
    ]
    """
    Provides all our known patterns.
    """

    REGEX_DIGITS: str = r"[0-9]"
    """
    Provides the regex to match in order to extract the digits.
    """

    MARKER2DATE_REGEX: dict = {
        # Date in format: 02-jan-2017
        "1": r"([0-9]{2})-([a-z]{3})-([0-9]{4})",
        # Date in format: 02.01.2017 // Month: jan
        "2": r"([0-9]{2})\.([0-9]{2})\.([0-9]{4})$",
        # Date in format: 02/01/2017 // Month: jan
        "3": r"([0-3][0-9])\/(0[1-9]|1[012])\/([0-9]{4})",
        # Date in format: 2017-01-02 // Month: jan
        "4": r"([0-9]{4})-([0-9]{2})-([0-9]{2})$",
        # Date in format: 2017.01.02 // Month: jan
        "5": r"([0-9]{4})\.([0-9]{2})\.([0-9]{2})$",
        # Date in format: 2017/01/02 // Month: jan
        "6": r"([0-9]{4})\/([0-9]{2})\/([0-9]{2})$",
        # Date in format: 2017.01.02 15:00:00
        "7": r"([0-9]{4})\.([0-9]{2})\.([0-9]{2})\s[0-9]{2}:[0-9]{2}:[0-9]{2}",
        # Date in format: 20170102 15:00:00 // Month: jan
        "8": r"([0-9]{4})([0-9]{2})([0-9]{2})\s[0-9]{2}:[0-9]{2}:[0-9]{2}",
        # Date in format: 2017-01-02 15:00:00 // Month: jan
        "9": r"([0-9]{4})-([0-9]{2})-([0-9]{2})\s[0-9]{2}:[0-9]{2}:[0-9]{2}",
        # Date in format: 02.01.2017 15:00:00 // Month: jan
        "10": r"([0-9]{2})\.([0-9]{2})\.([0-9]{4})\s[0-9]{2}:[0-9]{2}:[0-9]{2}",
        # Date in format: 02-Jan-2017 15:00:00 UTC
        "11": r"([0-9]{2})-([A-Z]{1}[a-z]{2})-([0-9]{4})\s[0-9]{2}:[0-9]{2}:[0-9]{2}\s[A-Z]{1}.*",  # pylint: disable=line-too-long  # noqa: E501
        # Date in format: 2017/01/02 01:00:00 (+0900) // Month: jan
        "12": r"([0-9]{4})\/([0-9]{2})\/([0-9]{2})\s[0-9]{2}:[0-9]{2}:[0-9]{2}\s\(.*\)",
        # Date in format: 2017/01/02 01:00:00 // Month: jan
        "13": r"([0-9]{4})\/([0-9]{2})\/([0-9]{2})\s[0-9]{2}:[0-9]{2}:[0-9]{2}$",
        # Date in format: Mon Jan 02 15:00:00 GMT 2017
        "14": r"[a-zA-Z]{3}\s([a-zA-Z]{3})\s([0-9]{2})\s[0-9]{2}:[0-9]{2}:[0-9]{2}\s[A-Z]{3}\s([0-9]{4})",  # pylint: disable=line-too-long  # noqa: E501
        # Date in format: Mon Jan 02 2017
        "15": r"[a-zA-Z]{3}\s([a-zA-Z]{3})\s([0-9]{2})\s([0-9]{4})",
        # Date in format: 2017-01-02T15:00:00 // Month: jan
        "16": r"([0-9]{4})-([0-9]{2})-([0-9]{2})T[0-9]{2}:[0-9]{2}:[0-9]{2}$",
        # Date in format: 2017-01-02T15:00:00Z // Month: jan${'7}
        "17": r"([0-9]{4})-([0-9]{2})-([0-9]{2})T[0-9]{2}:[0-9]{2}:[0-9]{2}[A-Z].*",
        # Date in format: 2017-01-02T15:00:00+0200 // Month: jan
        "18": r"([0-9]{4})-([0-9]{2})-([0-9]{2})T[0-9]{2}:[0-9]{2}:[0-9]{2}[+-][0-9]{4}",  # pylint: disable=line-too-long  # noqa: E501
        # Date in format: 2017-01-02T15:00:00+0200.622265+03:00 //
        # Month: jan
        "19": r"([0-9]{4})-([0-9]{2})-([0-9]{2})T[0-9]{2}:[0-9]{2}:[0-9]{2}\.[0-9].*[+-][0-9]{2}:[0-9]{2}",  # pylint: disable=line-too-long  # noqa: E501
        # Date in format: 2017-01-02T15:00:00+0200.622265 // Month: jan
        "20": r"([0-9]{4})-([0-9]{2})-([0-9]{2})T[0-9]{2}:[0-9]{2}:[0-9]{2}\.[0-9]{6}$",
        # Date in format: 2017-01-02T23:59:59.0Z // Month: jan
        "21": r"([0-9]{4})-([0-9]{2})-([0-9]{2})T[0-9]{2}:[0-9]{2}:[0-9]{2}\.[0-9].*[A-Z]",  # pylint: disable=line-too-long  # noqa: E501
        # Date in format: 02-01-2017 // Month: jan
        "22": r"([0-9]{2})-([0-9]{2})-([0-9]{4})",
        # Date in format: 2017. 01. 02. // Month: jan
        "23": r"([0-9]{4})\.\s([0-9]{2})\.\s([0-9]{2})\.",
        # Date in format: 2017-01-02T00:00:00+13:00 // Month: jan
        "24": r"([0-9]{4})-([0-9]{2})-([0-9]{2})T[0-9]{2}:[0-9]{2}:[0-9]{2}[+-][0-9]{2}:[0-9]{2}",  # pylint: disable=line-too-long  # noqa: E501
        # Date in format: 20170102 // Month: jan
        "25": r"(?=[0-9]{8})(?=([0-9]{4})([0-9]{2})([0-9]{2}))",
        # Date in format: 02-Jan-2017
        "26": r"([0-9]{2})-([A-Z]{1}[a-z]{2})-([0-9]{4})$",
        # Date in format: 02.1.2017 // Month: jan
        "27": r"([0-9]{2})\.([0-9]{1})\.([0-9]{4})",
        # Date in format: 02 Jan 2017
        "28": r"([0-9]{1,2})\s([A-Z]{1}[a-z]{2})\s([0-9]{4})",
        # Date in format: 02-January-2017
        "29": r"([0-9]{2})-([A-Z]{1}[a-z]*)-([0-9]{4})",
        # Date in format: 2017-Jan-02.
        "30": r"([0-9]{4})-([A-Z]{1}[a-z]{2})-([0-9]{2})\.",
        # Date in format: Mon Jan 02 15:00:00 2017
        "31": r"[a-zA-Z]{3}\s([a-zA-Z]{3})\s([0-9]{1,2})\s[0-9]{2}:[0-9]{2}:[0-9]{2}\s([0-9]{4})",  # pylint: disable=line-too-long   # noqa: E501
        # Date in format: Mon Jan 2017 15:00:00
        "32": r"()[a-zA-Z]{3}\s([a-zA-Z]{3})\s([0-9]{4})\s[0-9]{2}:[0-9]{2}:[0-9]{2}",
        # Date in format: January 02 2017-Jan-02
        "33": r"([A-Z]{1}[a-z]*)\s([0-9]{1,2})\s([0-9]{4})",
        # Date in format: 2.1.2017 // Month: jan
        "34": r"([0-9]{1,2})\.([0-9]{1,2})\.([0-9]{4})",
        # Date in format: 20170102000000 // Month: jan
        "35": r"([0-9]{4})([0-9]{2})([0-9]{2})[0-9]+",
        # Date in format: 01/02/2017 // Month: jan
        "36": r"(0[1-9]|1[012])\/([0-3][0-9])\/([0-9]{4})",
        # Date in format: January  2 2017
        "37": r"([A-Z]{1}[a-z].*)\s\s([0-9]{1,2})\s([0-9]{4})",
        # Date in format: 2nd January 2017
        "38": r"([0-9]{1,})[a-z]{1,}\s([A-Z].*)\s(2[0-9]{3})",
    }
    """
    Provides all the known regex to extract the date.

    .. note::
        The index (or key) will be used to map or group some of them later.
    """

    REGEX_PARSE_MAP: List[dict] = [
        {
            "regex_keys": [1, 2, 3, 10, 11, 22, 26, 27, 28, 29, 32, 34, 38],
            "positions": {"day": 0, "month": 1, "year": 2},
        },
        {
            "regex_keys": [14, 15, 31, 33, 36, 37],
            "positions": {"day": 1, "month": 0, "year": 2},
        },
        {
            "regex_keys": [
                4,
                5,
                6,
                7,
                8,
                9,
                12,
                13,
                16,
                17,
                18,
                19,
                20,
                21,
                23,
                24,
                25,
                30,
                35,
            ],
            "positions": {"day": 2, "month": 1, "year": 0},
        },
    ]
    """
    Our parsing map. Indeed, we hava a list of regex, but no way to know
    how to parse them. Especially when the order (month, day, year) are
    different from a format to another.

    This variable solve that problem by interpreting all regex we previously
    created.
    """

    @ConverterBase.data_to_convert.setter
    def data_to_convert(self, value: str) -> None:
        """
        Sets the data to convert or work with.

        :param value:
            The record to work with.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise ValueError:
            When the given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} is given.")

        if not value:
            raise ValueError("<value> should not be empty.")

        # pylint: disable=no-member
        super(ExpirationDateExtractor, self.__class__).data_to_convert.fset(self, value)

    def __get_line(self) -> Optional[str]:
        """
        Tries to get the expiration date line from the given record.
        """

        for regex in self.PATTERNS:
            expiration_date_line = RegexHelper(r"(?i)" + regex).match(
                self.data_to_convert, return_match=True, rematch=True, group=0
            )

            if not expiration_date_line:
                continue

            return expiration_date_line
        return None

    def __get_actual_expiration_date(
        self, extracted: str
    ) -> Optional[Tuple[str, str, str]]:
        """
        Tries to extract the actual expiration date.
        """

        for index, date_regex in self.MARKER2DATE_REGEX.items():
            matched = RegexHelper(date_regex).match(
                extracted, return_match=True, rematch=True
            )
            date_parts = tuple()

            if not matched:
                continue

            for parse_case in self.REGEX_PARSE_MAP:
                if int(index) not in parse_case["regex_keys"]:
                    continue

                date_parts = (
                    Digit2Digits(
                        matched[parse_case["positions"]["day"]],
                    ).get_converted(),
                    Month2Unified(
                        matched[parse_case["positions"]["month"]]
                    ).get_converted(),
                    str(matched[parse_case["positions"]["year"]]),
                )

            if date_parts:
                return "-".join(date_parts)
        return None

    @ConverterBase.ensure_data_to_convert_is_given
    def get_converted(self) -> Optional[str]:
        """
        Provides the expiration date of the record (if found).
        """

        expiration_date_line = self.__get_line()

        if expiration_date_line:
            expiration_date = expiration_date_line[0].strip()

            if RegexHelper(self.REGEX_DIGITS).match(
                expiration_date, return_match=False
            ):
                return self.__get_actual_expiration_date(expiration_date)

        return None



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the URL availability checker.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""


import PyFunceble.facility
import PyFunceble.factory
import PyFunceble.storage
from PyFunceble.checker.availability.base import AvailabilityCheckerBase
from PyFunceble.checker.availability.status import AvailabilityCheckerStatus
from PyFunceble.checker.reputation.url import URLReputationChecker
from PyFunceble.checker.syntax.url import URLSyntaxChecker


class URLAvailabilityChecker(AvailabilityCheckerBase):
    """
    Provides the interface for checking the availability of a given URL.

    :param str subject:
        Optional, The subject to work with.
    :param bool use_extra_rules:
        Optional, Activates/Disables the usage of our own set of extra rules.
    :param bool use_whois_lookup:
        Optional, Activates/Disables the usage of the WHOIS lookup to gather
        the status of the given :code:`subject`.
    :param bool use_dns_lookup:
        Optional, Activates/Disables the usage of the DNS lookup to gather the
        status of the given :code:`subject`.
    :param bool use_netinfo_lookup:
        Optional, Activates/Disables the usage of the network information
        lookup module to gather the status of the given :code:`subject`.
    :param bool use_http_code_lookup:
        Optional, Activates/Disables the usage of the HTTP status code lookup
        to gather the status of the given :code:`subject`.
    :param bool use_reputation_lookup:
        Optional, Activates/Disables the usage of the reputation dataset
        lookup to gather the status of the given :code:`subject`.
    :param bool do_syntax_check_first:
        Optional, Activates/Disables the check of the status before the actual
        status gathering.
    :param bool use_whois_db:
        Optional, Activates/Disable the usage of a local database to store the
        WHOIS datasets.
    """

    def subject_propagator(self) -> "URLAvailabilityChecker":
        """
        Propagate the currently set subject.

        .. warning::
            You are not invited to run this method directly.

        .. versionchanged:: 4.1.0b7.dev
           DNS Lookup capability.
        """

        self.http_status_code_query_tool.set_subject(self.idna_subject)
        self.dns_query_tool.set_subject(
            URLSyntaxChecker.get_hostname_from_url(self.idna_subject)
            or self.idna_subject
        )

        self.domain_syntax_checker.subject = self.idna_subject
        self.ip_syntax_checker.subject = self.idna_subject
        self.url_syntax_checker.subject = self.idna_subject

        self.status = AvailabilityCheckerStatus()
        self.status.params = self.params
        self.status.dns_lookup_record = self.dns_query_tool.lookup_record
        self.status.whois_lookup_record = None

        self.status.subject = self.subject
        self.status.idna_subject = self.idna_subject
        self.status.status = None

        self.query_syntax_checker()

        return self

    def try_to_query_status_from_http_status_code(self) -> "URLAvailabilityChecker":
        """
        Tries to query the status from the network information.
        """

        PyFunceble.facility.Logger.info(
            "Started to try to query the status of %r from: HTTP Status code Lookup",
            self.status.idna_subject,
        )

        lookup_result = self.http_status_code_query_tool.get_status_code()

        if (
            lookup_result
            and lookup_result
            != self.http_status_code_query_tool.STD_UNKNOWN_STATUS_CODE
        ):
            self.status.http_status_code = lookup_result

            if (
                PyFunceble.facility.ConfigLoader.is_already_loaded()
            ):  # pragma: no cover ## Special behavior
                dataset = PyFunceble.storage.HTTP_CODES
            else:
                dataset = PyFunceble.storage.STD_HTTP_CODES

            if (
                self.status.http_status_code in dataset.list.up
                or self.status.http_status_code in dataset.list.potentially_up
            ):
                self.status.status = PyFunceble.storage.STATUS.up
                self.status.status_source = "HTTP CODE"

                PyFunceble.facility.Logger.info(
                    "Could define the status of %r from: HTTP Status code Lookup",
                    self.status.idna_subject,
                )
        else:
            self.status.http_status_code = None

        PyFunceble.facility.Logger.info(
            "Finished to try to query the status of %r from: HTTP Status code Lookup",
            self.status.idna_subject,
        )

        return self

    def try_to_query_status_from_reputation(self) -> "URLAvailabilityChecker":
        """
        Tries to query the status from the reputation lookup.
        """

        PyFunceble.facility.Logger.info(
            "Started to try to query the status of %r from: Reputation Lookup",
            self.status.idna_subject,
        )

        lookup_result = URLReputationChecker(self.status.idna_subject).get_status()

        # pylint: disable=no-member
        if lookup_result and lookup_result.is_malicious():
            self.status.status = PyFunceble.storage.STATUS.up
            self.status.status_source = "REPUTATION"

            PyFunceble.facility.Logger.info(
                "Could define the status of %r from: Reputation Lookup",
                self.status.idna_subject,
            )

        PyFunceble.facility.Logger.info(
            "Started to try to query the status of %r from: Reputation Lookup",
            self.status.idna_subject,
        )

        return self

    def try_to_query_status_from_dns(self) -> "AvailabilityCheckerBase":
        """
        Tries to query the status from the DNS lookup after switching the
        idna subject to the url base.

        .. warning::
            This method does not answer as you may expect.

            Indeed, if a DNS lookup failed, this method will overwrite the
            standard response by setting the status to :code:`INACTIVE` and the
            status source to :code:`DNSLOOKUP`.

        .. versionadded:: 4.1.0b7
           DNS Lookup as a "down" switcher.
        """

        result = super().try_to_query_status_from_dns()

        if self.status.status == PyFunceble.storage.STATUS.up:
            # DNS should only be use to take subject down.
            # Therefore, switching back as if nothing happened.
            self.status.status = None
            self.status.status_source = None
        else:
            self.status.status = PyFunceble.storage.STATUS.down
            self.status.status_source = "DNSLOOKUP"

        return result

    @AvailabilityCheckerBase.ensure_subject_is_given
    @AvailabilityCheckerBase.update_status_date_after_query
    def query_status(
        self,
    ) -> "URLAvailabilityChecker":  # pragma: no cover
        """
        Queries the result without anything more.

        .. versionchanged:: 4.1.0b7.dev
            DNS Query - first.
        """

        ## Test Methods are more important.

        status_post_syntax_checker = None

        if not self.status.status and self.do_syntax_check_first:
            self.try_to_query_status_from_syntax_lookup()

            if self.status.status:
                status_post_syntax_checker = self.status.status

        if self.use_reputation_lookup and self.should_we_continue_test(
            status_post_syntax_checker
        ):
            self.try_to_query_status_from_reputation()

        if self.should_we_continue_test(status_post_syntax_checker):
            self.try_to_query_status_from_dns()

        if self.should_we_continue_test(status_post_syntax_checker):
            self.try_to_query_status_from_http_status_code()

        if not self.status.status:
            self.status.status = PyFunceble.storage.STATUS.down
            self.status.status_source = "STDLOOKUP"

            PyFunceble.facility.Logger.info(
                "Could not define status the status of %r. Setting to %r",
                self.status.idna_subject,
                self.status.status,
            )

        return self

    @staticmethod
    def is_valid() -> bool:  # pylint: disable=arguments-differ
        raise NotImplementedError()



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the file printer.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import functools
from datetime import datetime
from typing import Dict, Optional

import PyFunceble.facility
import PyFunceble.storage
from PyFunceble.cli.filesystem.printer.base import PrinterBase
from PyFunceble.helpers.file import FileHelper


class FilePrinter(PrinterBase):
    """
    Provides the file printer.
    """

    STD_FILE_GENERATION: str = (
        f"# Generated by {PyFunceble.storage.PROJECT_NAME} "
        f"(v{PyFunceble.storage.PROJECT_VERSION.split()[0]}) "
        f"/ {PyFunceble.storage.SHORT_REPO_LINK}\n"
    )

    file_helper: FileHelper = FileHelper()

    _destination: Optional[str] = None
    allow_coloration: bool = True

    def __init__(
        self,
        template_to_use: Optional[str] = None,
        *,
        dataset: Optional[Dict[str, str]] = None,
        destination: Optional[str] = None,
    ) -> None:
        if destination is not None:
            self.destination = destination

        super().__init__(template_to_use=template_to_use, dataset=dataset)

    def ensure_destination_is_given(func):  # pylint: disable=no-self-argument
        """
        Ensures that the destination is given before launching the
        decorated method.

        :raise TypeError:
            When the current :code:`self.template_to_use` is not set.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if not isinstance(self.destination, str):
                raise TypeError("<self.destination> is not given.")

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    def propagate_destination(func):  # pylint: disable=no-self-argument
        """
        Propagates the new value of the destination just after launching the
        decorated method.
        """

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            result = func(self, *args, **kwargs)  # pylint: disable=not-callable

            self.file_helper.set_path(self.destination)

            return result

        return wrapper

    @property
    def destination(self) -> Optional[str]:
        """
        Provides the current state of the :code:`_destination` attribute.
        """

        return self._destination

    @destination.setter
    @propagate_destination
    def destination(self, value: str) -> None:
        """
        Sets the destination to use.

        :param value:
            The value to set.

        :raise TypeError:
            When the given :code:`value` is not a :py:class:`str`.
        :raise ValueError:
            When teh given :code:`value` is empty.
        """

        if not isinstance(value, str):
            raise TypeError(f"<value> should be {str}, {type(value)} given.")

        if not value:
            raise ValueError("<value> should not be empty.")

        self._destination = value

    def set_destination(self, value: str) -> "FilePrinter":
        """
        Sets the destination to use.

        :param value:
            The value to set.
        """

        self.destination = value

        return self

    @staticmethod
    def get_generation_date_line() -> str:
        """
        Provides the line which informs of the date a file was generated.
        """

        return f"# Date of generation: {datetime.utcnow().isoformat()}"

    @ensure_destination_is_given
    def print_interpolated_line(self) -> None:
        """
        Prints the interpolated line into the destination.
        """

        PyFunceble.facility.Logger.info(
            "Started to write into %r.", self.file_helper.path
        )

        line_to_print = self.get_line_to_print() + "\n"
        without_header = ["hosts", "plain"]

        PyFunceble.facility.Logger.debug("Line to print: %r", line_to_print)

        if not self.file_helper.exists():
            self.file_helper.write(self.STD_FILE_GENERATION, overwrite=True)
            self.file_helper.write(self.get_generation_date_line())
            self.file_helper.write("\n\n")

            if self.template_to_use not in without_header:
                self.file_helper.write(self.get_header_to_print())
                self.file_helper.write("\n")

        self.file_helper.write(line_to_print)

        PyFunceble.facility.Logger.info(
            "Finished to write into %r.", self.file_helper.path
        )



"""
The tool to check the availability or syntax of domain, IP or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the migrator manager.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/#/special-thanks

Contributors:
    https://pyfunceble.github.io/#/contributors

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020, 2022 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from typing import Optional

import colorama
from sqlalchemy.orm import Session

import PyFunceble.checker.utils.whois
import PyFunceble.cli.utils.stdout
import PyFunceble.cli.utils.testing
import PyFunceble.facility
import PyFunceble.factory
import PyFunceble.storage
from PyFunceble.cli.continuous_integration.base import ContinuousIntegrationBase
from PyFunceble.cli.migrators.alembic import Alembic
from PyFunceble.cli.migrators.csv_file.inactive_source_delete import (
    InactiveDatasetDeleteSourceColumnMigrator,
)
from PyFunceble.cli.migrators.csv_file.whois_registrar_add import (
    WhoisDatasetAddRegistrarColumnMigrator,
)
from PyFunceble.cli.migrators.file_cleanup.hashes_file import HashesFileCleanupMigrator
from PyFunceble.cli.migrators.file_cleanup.mining_file import MiningFileCleanupMigrator
from PyFunceble.cli.migrators.file_cleanup.production_config_file import (
    ProductionConfigFileCleanupMigrator,
)
from PyFunceble.cli.migrators.json2csv.inactive import InactiveJSON2CSVMigrator
from PyFunceble.cli.migrators.json2csv.whois import WhoisJSON2CSVMigrator
from PyFunceble.cli.migrators.mariadb.file_and_status import FileAndStatusMigrator
from PyFunceble.cli.migrators.mariadb.whois_record_idna_subject import (
    WhoisRecordIDNASubjectMigrator,
)
from PyFunceble.cli.processes.base import ProcessesManagerBase
from PyFunceble.cli.processes.workers.migrator import MigratorWorker
from PyFunceble.helpers.file import FileHelper


class MigratorProcessesManager(ProcessesManagerBase):
    """
    Provides the migrator manager.
    """

    WORKER_OBJ: MigratorWorker = MigratorWorker

    @staticmethod
    def json2csv_inactive_target(
        continuous_integration: ContinuousIntegrationBase,
    ) -> None:
        """
        Provides the target for the inactive database migrator.
        """

        migrator = InactiveJSON2CSVMigrator(print_action_to_stdout=True)
        migrator.continuous_integration = continuous_integration

        if FileHelper(migrator.source_file).exists():
            print(
                f"{colorama.Fore.MAGENTA}{colorama.Style.BRIGHT}"
                "Started migration (json2csv) of the inactive dataset."
            )

            migrator.start()

            if migrator.done:
                print(
                    f"\n{colorama.Fore.GREEN}{colorama.Style.BRIGHT}"
                    "Finished migration (json2csv) of the inactive dataset."
                )
            else:
                print(
                    f"\n{colorama.Fore.MAGENTA}{colorama.Style.BRIGHT}"
                    "Unfinished migration (json2csv) of the inactive dataset."
                )
        else:
            PyFunceble.facility.Logger.info(
                "Stopped json2csv_inactive_target. File does not exist."
            )

    @staticmethod
    def json2csv_whois_target(
        continuous_integration: ContinuousIntegrationBase,
    ) -> None:
        """
        Provides the target for the whois database migrator.
        """

        migrator = WhoisJSON2CSVMigrator(print_action_to_stdout=True)
        migrator.continuous_integration = continuous_integration

        if FileHelper(migrator.source_file).exists():
            print(
                f"{colorama.Fore.MAGENTA}{colorama.Style.BRIGHT}"
                "Started migration (json2csv) of the whois dataset."
            )

            migrator.start()

            if migrator.done:
                print(
                    f"\n{colorama.Fore.GREEN}{colorama.Style.BRIGHT}"
                    "Finished migration (json2csv) of the whois dataset."
                )
            else:
                print(
                    f"\n{colorama.Fore.MAGENTA}{colorama.Style.BRIGHT}"
                    "Unfinished migration (json2csv) of the whois dataset."
                )
        else:
            PyFunceble.facility.Logger.info(
                "Stopped json2csv_whois_target. File does not exist."
            )

    @staticmethod
    def mariadb_whois_record_idna_subject_target(
        continuous_integration: ContinuousIntegrationBase,
        *,
        db_session: Optional[Session] = None,
    ) -> None:
        """
        Provides the target for the whois addition of the missing
        idna_subject column.
        """

        migrator = WhoisRecordIDNASubjectMigrator(print_action_to_stdout=True)
        migrator.continuous_integration = continuous_integration
        migrator.db_session = db_session

        if migrator.authorized:
            print(
                f"{colorama.Fore.MAGENTA}{colorama.Style.BRIGHT}"
                "Started completion of NULL idna_subject(s) into the whois dataset."
            )

            migrator.start()

            if migrator.done:
                print(
                    f"{colorama.Fore.GREEN}{colorama.Style.BRIGHT}"
                    "Finished completion of NULL idna_subject(s) into "
                    "the whois dataset."
                )
            else:
                print(
                    f"{colorama.Fore.MAGENTA}{colorama.Style.BRIGHT}"
                    "Unfinished completion of NULL idna_subject(s) into "
                    "the whois dataset."
                )
        else:
            PyFunceble.facility.Logger.info(
                "Stopped mariadb_whois_record_idna_subject_target. Not authorized."
            )

    @staticmethod
    def mariadb_file_and_status_target(
        continuous_integration: ContinuousIntegrationBase,
        *,
        db_session: Optional[Session] = None,
    ) -> None:
        """
        Provides the target for the migration of the :code:`pyfunceble_file`
        and :code:`pyfunceble_status` tables.
        """

        migrator = FileAndStatusMigrator(print_action_to_stdout=True)
        migrator.continuous_integration = continuous_integration
        migrator.db_session = db_session

        if migrator.authorized:
            print(
                f"{colorama.Fore.MAGENTA}{colorama.Style.BRIGHT}"
                "Started migration of the pyfunceble_file and "
                "pyfunceble_status tables."
            )

            migrator.start()

            if migrator.done:
                print(
                    f"{colorama.Fore.GREEN}{colorama.Style.BRIGHT}"
                    "Finished migration of the pyfunceble_file and "
                    "pyfunceble_status tables."
                )
            else:
                print(
                    f"{colorama.Fore.MAGENTA}{colorama.Style.BRIGHT}"
                    "Unfinished migration of the pyfunceble_file and "
                    "pyfunceble_status tables."
                )
        else:
            PyFunceble.facility.Logger.info(
                "Stopped mariadb_file_and_status_target. Not authorized."
            )

    @staticmethod
    def hashes_file_cleanup_target(
        continuous_integration: ContinuousIntegrationBase,
    ) -> None:
        """
        Provides the target for the cleanup of the hashes file.
        """

        migrator = HashesFileCleanupMigrator(print_action_to_stdout=True)
        migrator.continuous_integration = continuous_integration

        if FileHelper(migrator.source_file).exists():
            print(
                f"{colorama.Fore.MAGENTA}{colorama.Style.BRIGHT}"
                f"Started deletion of {migrator.source_file!r}."
            )

            migrator.start()

            if migrator.done:
                print(
                    f"{colorama.Fore.GREEN}{colorama.Style.BRIGHT}"
                    f"Finished deletion of {migrator.source_file!r}."
                )
            else:
                print(
                    f"{colorama.Fore.MAGENTA}{colorama.Style.BRIGHT}"
                    f"Unfinished deletion of {migrator.source_file!r}."
                )
        else:
            PyFunceble.facility.Logger.info(
                "Stopped hashes_file_cleanup_target. File does not exist."
            )

    @staticmethod
    def production_config_file_cleanup_target(
        continuous_integration: ContinuousIntegrationBase,
    ) -> None:
        """
        Provides the target for the cleanup of the production configuration file.
        """

        migrator = ProductionConfigFileCleanupMigrator(print_action_to_stdout=True)
        migrator.continuous_integration = continuous_integration

        if FileHelper(migrator.source_file).exists():
            print(
                f"{colorama.Fore.MAGENTA}{colorama.Style.BRIGHT}"
                f"Started deletion of {migrator.source_file!r}."
            )

            migrator.start()

            if migrator.done:
                print(
                    f"{colorama.Fore.GREEN}{colorama.Style.BRIGHT}"
                    f"Finished deletion of {migrator.source_file!r}."
                )
            else:
                print(
                    f"{colorama.Fore.MAGENTA}{colorama.Style.BRIGHT}"
                    f"Unfinished deletion of {migrator.source_file!r}."
                )
        else:
            PyFunceble.facility.Logger.info(
                "Stopped production_config_file_cleanup_target. File does not exist."
            )

    @staticmethod
    def mining_file_cleanup_target(
        continuous_integration: ContinuousIntegrationBase,
    ) -> None:
        """
        Provides the target for the cleanup of the mining file.
        """

        migrator = MiningFileCleanupMigrator(print_action_to_stdout=True)
        migrator.continuous_integration = continuous_integration

        if FileHelper(migrator.source_file).exists():
            print(
                f"{colorama.Fore.MAGENTA}{colorama.Style.BRIGHT}"
                f"Started deletion of {migrator.source_file!r}."
            )

            migrator.start()

            if migrator.done:
                print(
                    f"{colorama.Fore.GREEN}{colorama.Style.BRIGHT}"
                    f"Finished deletion of {migrator.source_file!r}."
                )
            else:
                print(
                    f"{colorama.Fore.MAGENTA}{colorama.Style.BRIGHT}"
                    f"Unfinished deletion of {migrator.source_file!r}."
                )
        else:
            PyFunceble.facility.Logger.info(
                "Stopped hashes_file_cleanup_target. File does not exist."
            )

    @staticmethod
    def csv_file_delete_source_column_target(
        continuous_integration: ContinuousIntegrationBase,
    ) -> None:
        """
        Provides the target for the deletion of the source column.
        """

        migrator = InactiveDatasetDeleteSourceColumnMigrator(
            print_action_to_stdout=True
        )
        migrator.continuous_integration = continuous_integration

        file_helper = FileHelper(migrator.source_file)

        if file_helper.exists():
            with file_helper.open("r", encoding="utf-8") as file_stream:
                first_line = next(file_stream)

            if any(x in first_line for x in migrator.TO_DELETE):
                print(
                    f"{colorama.Fore.MAGENTA}{colorama.Style.BRIGHT}"
                    "Started deletion of the 'source' column into "
                    f"{migrator.source_file!r}."
                )

                migrator.start()

                if migrator.done:
                    print(
                        f"{colorama.Fore.GREEN}{colorama.Style.BRIGHT}"
                        "Finished deletion of the 'source' column into "
                        f"{migrator.source_file!r}."
                    )
                else:
                    print(
                        f"{colorama.Fore.MAGENTA}{colorama.Style.BRIGHT}"
                        "unfinished deletion of the 'source' column into "
                        f"{migrator.source_file!r}."
                    )
        else:
            PyFunceble.facility.Logger.info(
                "Stopped csv_file_delete_source_column_target. File does not exist."
            )

    @staticmethod
    def csv_file_add_registrar_column_target(
        continuous_integration: ContinuousIntegrationBase,
    ) -> None:
        """
        Provides the target for the addition of the registrar column.
        """

        migrator = WhoisDatasetAddRegistrarColumnMigrator(print_action_to_stdout=True)
        migrator.continuous_integration = continuous_integration

        file_helper = FileHelper(migrator.source_file)

        if file_helper.exists():
            with file_helper.open("r", encoding="utf-8") as file_stream:
                first_line = next(file_stream)

            if any(x not in first_line for x in migrator.TO_ADD):
                print(
                    f"{colorama.Fore.MAGENTA}{colorama.Style.BRIGHT}"
                    "Started addition of the 'registrar' column into "
                    f"{migrator.source_file!r}."
                )

                migrator.start()

                if migrator.done:
                    print(
                        f"{colorama.Fore.GREEN}{colorama.Style.BRIGHT}"
                        "Finished addition of the 'registrar' column into "
                        f"{migrator.source_file!r}."
                    )
                else:
                    print(
                        f"{colorama.Fore.MAGENTA}{colorama.Style.BRIGHT}"
                        "unfinished addition of the 'registrar' column into "
                        f"{migrator.source_file!r}."
                    )
        else:
            PyFunceble.facility.Logger.info(
                "Stopped csv_file_add_registrar_column_target. File does not exist."
            )

    def create(self) -> "ProcessesManagerBase":
        for method in dir(self):
            if not method.endswith("_target"):
                continue

            worker = MigratorWorker(
                None,
                name=f"pyfunceble_{method}",
                daemon=True,
                continuous_integration=self.continuous_integration,
            )

            worker.target = getattr(self, method)

            self._created_workers.append(worker)
            PyFunceble.facility.Logger.info("Created worker for %r", method)

    @ProcessesManagerBase.ensure_worker_obj_is_given
    @ProcessesManagerBase.create_workers_if_missing
    def start(self) -> "ProcessesManagerBase":
        # We start the migration (as a standalone)
        Alembic(self._created_workers[0].db_session).upgrade()

        return super().start()
